from datetime import datetime

from persistence import Connection

from application.model import (
    UpdateStatus,
    UpdateType,
    UserUpdateData,
)
from application.model.configuration import Configurations
from application.model.customexception import ErrorInsertingNewUpdateVersion
from boss_api import Boss
from boss_api.shared.response import UserDto
from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from typing import (
    List,
    Optional,
)
from ._CommonPOSRepository import CommonPOSRepository


class UserUpdaterPOSRepository(BaseRepository):

    def __init__(
        self,           # type: UserUpdaterPOSRepository
        message_bus,    # type: MbContextMessageBus
        configs,        # type: Configurations
    ):
        # type: (...) -> None

        super(UserUpdaterPOSRepository, self).__init__(message_bus.mbcontext)

        self.message_bus = message_bus
        self.configs = configs
        self.update_type = UpdateType.user
        self.update_value = self.update_type.value
        self.local_configs = configs.updaters[self.update_type.name]
        self.common_pos_repository = CommonPOSRepository(message_bus)
        self.boss_service = Boss(message_bus=message_bus)

    def get_pos_users(
        self,   # type: UserUpdaterPOSRepository
    ):
        # type: (...) -> List[UserDto]

        return self.boss_service.get_users_list()

    def update_users(
        self,           # type: UserUpdaterPOSRepository
        updated_users,  # type: List[UserUpdateData]
    ):
        # type: (...) -> None

        updated_users = [user.to_boss_user_dto() for user in updated_users]
        self.boss_service.insert_or_update_users(users=updated_users)

    def insert_new_user_update(self, update_id, update_name):
        # type: (str, str) -> None

        try:
            def inner_func(conn):
                # type: (Connection) -> None

                date_now = str(datetime.now().isoformat())
                query = """
                INSERT INTO UpdatesController
                (UpdateId, UpdateName, TypeId, StatusId, ObtainedDate, DownloadedDate, BackupDate, AppliedDate,
                NotifiedDate)
                VALUES ('{0}', '{1}', '{2}', '{3}', '{4}', '{4}', '{4}', '{4}', '{4}')"""
                query_format = query.format(update_id, update_name, self.update_value, UpdateStatus.applied, date_now)
                conn.query(query_format)

            self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)
        except Exception as ex:
            raise ErrorInsertingNewUpdateVersion(ex)

    def get_last_user_update_id(self):
        # type: () -> int

        def inner_func(conn):
            # type: (Connection) -> int

            query = """
                SELECT UpdateId
                FROM UpdatesController
                WHERE TypeId = '{}' AND AppliedDate IS NOT NULL
                ORDER BY Id DESC
                LIMIT 1""".format(self.update_value)

            cursor = conn.select(query)
            row = cursor.get_row(0)

            if row:
                update_id = row.get_entry("UpdateId")
                return int(update_id)

            return 0

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def has_already_applied_update(self):
        # type: () -> bool

        return self.common_pos_repository.has_already_applied_update(self.configs.persistence_name, self.update_value)

    def disable_users(
        self,           # type: UserUpdaterPOSRepository
        users_disable,  # type: List[str]
    ):
        # type: (...) -> None

        users = self.boss_service.get_users_list()
        for user in users:
            if str(user.user_id) in users_disable:
                user.status = 1
                user.termination_date = datetime.now().date().isoformat()

        users_to_disable = [user for user in users if user.status == 1]
        self.boss_service.insert_or_update_users(users=users_to_disable)

    def get_last_user_update_date(self):
        # type: () -> Optional[datetime]

        return self.common_pos_repository.get_last_update_date(self.configs.persistence_name, self.update_value)



import logging
from typing import (
    List,
    Optional,
    Tuple,
)
from urllib.parse import urljoin

import requests
from boss.model.service import AuthenticationService
from boss.model.service.dto import AuthenticationData
from boss_api.shared.exception import UnexpectedError
from helper import retry
from message_handler_builder import Configs
from requests import Response

logger = logging.getLogger(__name__)


class AuthenticateServiceImpl(AuthenticationService):

    def __init__(
        self,
        configs: Configs,
    ) -> None:

        self.configs = configs

    def authenticate(
        self,
        username: int,
        password: str,
    ) -> Optional[AuthenticationData]:

        try:
            body = dict(
                username=str(username),
                password=str(password),
            )
            errors_list = []

            response = retry(
                logger=logger,
                method_name="authenticate",
                n=self.configs.api_retry_policy.count,
                sleep_seconds=self.configs.api_retry_policy.sleep,
                trying_to_do=lambda: requests.request(
                    method="POST",
                    url=self._get_url(),
                    json=body,
                    headers={
                        "x-api-key": self.configs.pump_credentials.api_key,
                        "Content-Type": "application/json",
                    },
                    timeout=5,
                ),
                is_success=lambda resp: resp.status_code == 200,
                success=lambda resp: resp,
                failed=lambda resp: errors_list.append((resp.status_code, resp)),
                on_exception=lambda exception: errors_list.append(exception)
            )
            if response is not None:
                response_data = response.json()
                return AuthenticationData(
                    generate_by=response_data["generateBy"],
                    user_id=response_data["userId"],
                    user_name=response_data["userName"],
                    long_name=response_data["longName"],
                    user_level=response_data["userLevel"],
                    authentication_key=response_data["authenticationKey"],
                    generation_date=response_data["generationDate"],
                    expiration_date=response_data["expirationDate"],
                    interval=response_data["interval"],
                )

            response_error = self._get_response_error(errors_list=errors_list)

            logger.info(f"Fail on authenticate api request: {response_error.status_code} - {response_error.text}")

            if response_error.status_code in [401, 403, 404]:
                return None

            unexpected_error = UnexpectedError()
            unexpected_error.error_message = response_error.text
            raise unexpected_error
        except Exception as ex:
            logger.error(f"Error on authenticate api request: {str(ex)}")
            raise

    def _get_url(
        self,
    ) -> str:

        endpoint = self.configs.endpoints.authenticate.format(store_code=self.configs.store_id)
        return urljoin(
            base=self.configs.pump_credentials.base_url.rstrip("/"),
            url=endpoint.lstrip("/"),
        )

    @staticmethod
    def _get_response_error(
        errors_list: List[Tuple[int, Response]],
    ) -> Optional[Response]:

        if not errors_list:
            return None

        for last_error in reversed(errors_list):
            if not isinstance(last_error, Exception):
                _, response = last_error
                return response

        raise errors_list[-1]



from bustoken import TK_REMOTE_ORDER_GET_POS_ID
from msgbus import (
    FM_PARAM,
    TK_SYS_NAK,
)
from sysactions import send_message
from typing import Optional
from xml.etree import cElementTree as eTree

from api.order import get_order_service

pos_id_remote_order = None


def is_delivery_order(
    pos_id,         # type: str
    order=None,     # type: eTree.Element
):
    # type: (...) -> bool

    order = order or get_order_service(pos_id=pos_id).get_order_picture()
    if order is None:
        return False

    has_remote_order_id = get_remote_id(order=order) is not None
    return has_remote_order_id


def get_remote_id(
    order,  # type: eTree.Element
):
    # type: (...) -> Optional[str]

    remote_order_id = order.find("CustomOrderProperties/OrderProperty[@key='REMOTE_ORDER_ID']")
    if remote_order_id is not None:
        return remote_order_id.get("value")

    manual_order_id = order.find("CustomOrderProperties/OrderProperty[@key='MANUAL_ORDER_ID']")
    if manual_order_id is not None:
        return manual_order_id.get("value")

    return None


def get_pos_id_of_remote_order():
    # type: (...) -> int

    global pos_id_remote_order

    if pos_id_remote_order is not None:
        return pos_id_remote_order

    msg = send_message(
        dest_name="RemoteOrder",
        token=TK_REMOTE_ORDER_GET_POS_ID,
        format=FM_PARAM,
        timeout=100000,
    )
    if msg.token == TK_SYS_NAK:
        return 0

    pos_id_remote_order = int(msg.data)
    return pos_id_remote_order



from sqlite3 import Connection
from typing import Any

from messageprocessor import MessageProcessorCallback


class SqliteConnectionTransactionCallback(MessageProcessorCallback):

    def __init__(
        self,
        conn: Connection,
    ) -> None:

        self.conn = conn

    def process_begun(
        self,
        event_name: str,
        request_id: str,
        **kwargs
    ) -> None:

        return

    def input_parsed(
        self,
        event_name: str,
        request_id: str,
        input_obj: Any,
    ) -> None:

        return

    def business_called(
        self,
        event_name: str,
        request_id: str,
        input_obj: Any,
        output_obj: Any,
    ) -> None:

        if self.conn.total_changes > 0:
            self.conn.commit()

    def process_finished(
        self,
        event_name: str,
        request_id: str,
        input_obj: Any,
        output_obj: Any,
    ) -> None:

        return

    def parse_exception(
        self,
        event_name: str,
        request_id: str,
        data: bytes,
        exception: Exception,
    ) -> None:

        return

    def unhandled_exception(
        self,
        event_name: str,
        request_id: str,
        input_obj: Any,
        exception: Exception,
    ) -> None:

        if self.conn.total_changes > 0:
            self.conn.commit()
        else:
            self.conn.rollback()



# coding: utf-8
"""
Sphinx directive integration
============================

We usually need to document the life-cycle of functions and classes:
when they are created, modified or deprecated.

To do that, `Sphinx <http://www.sphinx-doc.org>`_ has a set
of `Paragraph-level markups <http://www.sphinx-doc.org/en/stable/markup/para.html>`_:

- ``versionadded``: to document the version of the project which added the described feature to the library,
- ``versionchanged``: to document changes of a feature,
- ``deprecated``: to document a deprecated feature.

The purpose of this module is to defined decorators which adds this Sphinx directives
to the docstring of your function and classes.

Of course, the ``@deprecated`` decorator will emit a deprecation warning
when the function/method is called or the class is constructed.
"""
import re
import textwrap

from deprecated.classic import ClassicAdapter
from deprecated.classic import deprecated as _classic_deprecated


class SphinxAdapter(ClassicAdapter):
    """
    Sphinx adapter -- *for advanced usage only*

    This adapter override the :class:`~deprecated.classic.ClassicAdapter`
    in order to add the Sphinx directives to the end of the function/class docstring.
    Such a directive is a `Paragraph-level markup <http://www.sphinx-doc.org/en/stable/markup/para.html>`_

    - The directive can be one of "versionadded", "versionchanged" or "deprecated".
    - The version number is added if provided.
    - The reason message is obviously added in the directive block if not empty.
    """

    def __init__(
        self,
        directive,
        reason="",
        version="",
        action=None,
        category=DeprecationWarning,
        extra_stacklevel=0,
        line_length=70,
    ):
        """
        Construct a wrapper adapter.

        :type  directive: str
        :param directive:
            Sphinx directive: can be one of "versionadded", "versionchanged" or "deprecated".

        :type  reason: str
        :param reason:
            Reason message which documents the deprecation in your library (can be omitted).

        :type  version: str
        :param version:
            Version of your project which deprecates this feature.
            If you follow the `Semantic Versioning <https://semver.org/>`_,
            the version number has the format "MAJOR.MINOR.PATCH".

        :type  action: Literal["default", "error", "ignore", "always", "module", "once"]
        :param action:
            A warning filter used to activate or not the deprecation warning.
            Can be one of "error", "ignore", "always", "default", "module", or "once".
            If ``None`` or empty, the global filtering mechanism is used.
            See: `The Warnings Filter`_ in the Python documentation.

        :type  category: Type[Warning]
        :param category:
            The warning category to use for the deprecation warning.
            By default, the category class is :class:`~DeprecationWarning`,
            you can inherit this class to define your own deprecation warning category.

        :type  extra_stacklevel: int
        :param extra_stacklevel:
            Number of additional stack levels to consider instrumentation rather than user code.
            With the default value of 0, the warning refers to where the class was instantiated
            or the function was called.

        :type  line_length: int
        :param line_length:
            Max line length of the directive text. If non nul, a long text is wrapped in several lines.

        .. versionchanged:: 1.2.15
            Add the *extra_stacklevel* parameter.
        """
        if not version:
            # https://github.com/laurent-laporte-pro/deprecated/issues/40
            raise ValueError("'version' argument is required in Sphinx directives")
        self.directive = directive
        self.line_length = line_length
        super(SphinxAdapter, self).__init__(
            reason=reason, version=version, action=action, category=category, extra_stacklevel=extra_stacklevel
        )

    def __call__(self, wrapped):
        """
        Add the Sphinx directive to your class or function.

        :param wrapped: Wrapped class or function.

        :return: the decorated class or function.
        """
        # -- build the directive division
        fmt = ".. {directive}:: {version}" if self.version else ".. {directive}::"
        div_lines = [fmt.format(directive=self.directive, version=self.version)]
        width = self.line_length - 3 if self.line_length > 3 else 2**16
        reason = textwrap.dedent(self.reason).strip()
        for paragraph in reason.splitlines():
            if paragraph:
                div_lines.extend(
                    textwrap.fill(
                        paragraph,
                        width=width,
                        initial_indent="   ",
                        subsequent_indent="   ",
                    ).splitlines()
                )
            else:
                div_lines.append("")

        # -- get the docstring, normalize the trailing newlines
        # keep a consistent behaviour if the docstring starts with newline or directly on the first one
        docstring = wrapped.__doc__ or ""
        lines = docstring.splitlines(True) or [""]
        docstring = textwrap.dedent("".join(lines[1:])) if len(lines) > 1 else ""
        docstring = lines[0] + docstring
        if docstring:
            # An empty line must separate the original docstring and the directive.
            docstring = re.sub(r"\n+$", "", docstring, flags=re.DOTALL) + "\n\n"
        else:
            # Avoid "Explicit markup ends without a blank line" when the decorated function has no docstring
            docstring = "\n"

        # -- append the directive division to the docstring
        docstring += "".join("{}\n".format(line) for line in div_lines)

        wrapped.__doc__ = docstring
        if self.directive in {"versionadded", "versionchanged"}:
            return wrapped
        return super(SphinxAdapter, self).__call__(wrapped)

    def get_deprecated_msg(self, wrapped, instance):
        """
        Get the deprecation warning message (without Sphinx cross-referencing syntax) for the user.

        :param wrapped: Wrapped class or function.

        :param instance: The object to which the wrapped function was bound when it was called.

        :return: The warning message.

        .. versionadded:: 1.2.12
           Strip Sphinx cross-referencing syntax from warning message.

        """
        msg = super(SphinxAdapter, self).get_deprecated_msg(wrapped, instance)
        # Strip Sphinx cross-reference syntax (like ":function:", ":py:func:" and ":py:meth:")
        # Possible values are ":role:`foo`", ":domain:role:`foo`"
        # where ``role`` and ``domain`` should match "[a-zA-Z]+"
        msg = re.sub(r"(?: : [a-zA-Z]+ )? : [a-zA-Z]+ : (`[^`]*`)", r"\1", msg, flags=re.X)
        return msg


def versionadded(reason="", version="", line_length=70):
    """
    This decorator can be used to insert a "versionadded" directive
    in your function/class docstring in order to document the
    version of the project which adds this new functionality in your library.

    :param str reason:
        Reason message which documents the addition in your library (can be omitted).

    :param str version:
        Version of your project which adds this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH", and,
        in the case of a new functionality, the "PATCH" component should be "0".

    :type  line_length: int
    :param line_length:
        Max line length of the directive text. If non nul, a long text is wrapped in several lines.

    :return: the decorated function.
    """
    adapter = SphinxAdapter(
        'versionadded',
        reason=reason,
        version=version,
        line_length=line_length,
    )
    return adapter


def versionchanged(reason="", version="", line_length=70):
    """
    This decorator can be used to insert a "versionchanged" directive
    in your function/class docstring in order to document the
    version of the project which modifies this functionality in your library.

    :param str reason:
        Reason message which documents the modification in your library (can be omitted).

    :param str version:
        Version of your project which modifies this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH".

    :type  line_length: int
    :param line_length:
        Max line length of the directive text. If non nul, a long text is wrapped in several lines.

    :return: the decorated function.
    """
    adapter = SphinxAdapter(
        'versionchanged',
        reason=reason,
        version=version,
        line_length=line_length,
    )
    return adapter


def deprecated(reason="", version="", line_length=70, **kwargs):
    """
    This decorator can be used to insert a "deprecated" directive
    in your function/class docstring in order to document the
    version of the project which deprecates this functionality in your library.

    :param str reason:
        Reason message which documents the deprecation in your library (can be omitted).

    :param str version:
        Version of your project which deprecates this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH".

    :type  line_length: int
    :param line_length:
        Max line length of the directive text. If non nul, a long text is wrapped in several lines.

    Keyword arguments can be:

    -   "action":
        A warning filter used to activate or not the deprecation warning.
        Can be one of "error", "ignore", "always", "default", "module", or "once".
        If ``None``, empty or missing, the global filtering mechanism is used.

    -   "category":
        The warning category to use for the deprecation warning.
        By default, the category class is :class:`~DeprecationWarning`,
        you can inherit this class to define your own deprecation warning category.

    -   "extra_stacklevel":
        Number of additional stack levels to consider instrumentation rather than user code.
        With the default value of 0, the warning refers to where the class was instantiated
        or the function was called.


    :return: a decorator used to deprecate a function.

    .. versionchanged:: 1.2.13
       Change the signature of the decorator to reflect the valid use cases.

    .. versionchanged:: 1.2.15
        Add the *extra_stacklevel* parameter.
    """
    directive = kwargs.pop('directive', 'deprecated')
    adapter_cls = kwargs.pop('adapter_cls', SphinxAdapter)
    kwargs["reason"] = reason
    kwargs["version"] = version
    kwargs["line_length"] = line_length
    return _classic_deprecated(directive=directive, adapter_cls=adapter_cls, **kwargs)



import os

from config import Configurator
from edplogutil import wrap_types
from helper import (
    import_pydevd,
)
from mbcontextmessagehandler import (
    MbContextMessageHandler,
    MbContextMessageBus,
)
from message_handler_builder import (
    EventManagerMessageHandlerBuilder,
)
from msgbus import MBEasyContext

# from opentelemetry_edphelper import configure_open_telemetry

from poshelper import config_logger_parse_file_name
from sqlite_repository import SqliteEventRepository
from timeutil import (
    RealClock,
    ConditionSleeper,
)


def main():
    loader_path = os.environ["LOADERCFG"]

    import_pydevd(loader_path, 9145)

    # configure_open_telemetry(
    #     service_name="EventManager",
    # )

    wrap_types(
        clss=[
            RealClock,
            ConditionSleeper,
            MbContextMessageBus,
            SqliteEventRepository,
        ]
    )

    message_bus = MbContextMessageBus(mbcontext=MBEasyContext("CustomEventManager"))

    config_logger_parse_file_name(os.environ["LOG_PATH"])
    config_info = Configurator(loader_path).read_loader_configs()

    message_handler_builder = EventManagerMessageHandlerBuilder(
        config_info=config_info,
        message_bus=message_bus,
        clock=RealClock(),
    )
    message_handler = MbContextMessageHandler(
        message_bus=message_bus,
        service_name=config_info.component_name,
        service_type=config_info.component_name,
        required_services="",
        event_handler_builder=message_handler_builder
    )

    return message_handler.handle_events()



INVALID_JSON = 0
UNEXPECTED_ERROR = -1
MISSING_REQUEST_NAME = -2
REPORT_FILE_NOT_FOUND = -3
REPORT_NOT_IMPLEMENTED = -4
ERROR_BUILDING_REPORT_DATA = -5
ERROR_BUILDING_DATA_BUILDER = -6
ERROR_FORMATTING_REPORT = -7
REPORT_FUNCTION_NOT_IMPLEMENTED = -8
INVALID_TEXT_FORMAT_PARAMETERS_COUNT = -9
TEXT_FORMAT_LENGTH_NOT_INTEGER = -10
INVALID_TEXT_FORMAT_SIDE_PARAMETER = -11
INVALID_CURRENCY_PARAMETERS_COUNT = -12
CURRENCY_VALUE_NOT_A_NUMBER = -13
INVALID_NUMBER_PARAMETERS_COUNT = -14
NUMBER_VALUE_NOT_A_NUMBER = -15
DATETIME_NOW_CANNOT_RECEIVE_PARAMETER = -16
INVALID_DATETIME_PARAMETERS_COUNT = -17
DATETIME_DATE_NOT_A_ISO_DATE = -18
I18N_LABEL_NOT_FOUND = -19
COULD_NOT_LOAD_POS_TRANSLATIONS = -20
CONDITION_IS_NOT_A_VARIABLE = -21
CONDITION_VARIABLE_NOT_PRESENT = -22
VALUE_RETURNED_FROM_OBJECT_LIST_IS_NOT_A_LIST = -23
VARIABLE_WAS_NOT_GENERATED_BY_THE_BUILDER = -24
TEXT_LIST_ATTRIBUTE_IS_NOT_A_LIST = -25
PADDING_IS_NOT_A_NUMBER = -26
INNER_LIST_ATTRIBUTE_IS_NOT_A_LIST = -27
INNER_LIST_MUST_HAVE_NAME = -28
ERROR_PARSING_REPORT_DATA = -29



import os

import cfgtools

is_change_fund_config_enabled = None


def is_change_fund_enabled():
    # type: (...) -> bool

    global is_change_fund_config_enabled

    if is_change_fund_config_enabled is None:
        config = cfgtools.read(file_or_path=os.environ["LOADERCFG"])
        is_change_fund_config_enabled = config.find_value(
            path="Account.EnableChangeFund",
            default="false",
        )

    return is_change_fund_config_enabled.lower() == "true"



# vnr.py - functions for handling Austrian social security numbers
# coding: utf-8
#
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""VNR, SVNR, VSNR (Versicherungsnummer, Austrian social security number).

The Austian Versicherungsnummer is a personal identification number used for
social security. The number is 10 digits long and consists of a 3 digit
serial, a check digit and 6 digits that usually specify the person's birth
date.

More information:

* https://de.wikipedia.org/wiki/Sozialversicherungsnummer#Österreich

>>> validate('1237 010180')
'1237010180'
>>> validate('2237 010180')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ')


def calc_check_digit(number):
    """Calculate the check digit. The fourth digit in the number is
    ignored."""
    weights = (3, 7, 9, 0, 5, 8, 4, 2, 1, 6)
    return str(sum(w * int(n) for w, n in zip(weights, number)) % 11)


def validate(number):
    """Check if the number is a valid VAT number. This checks the length,
    formatting and check digit."""
    number = compact(number)
    if not isdigits(number) or number.startswith('0'):
        raise InvalidFormat()
    if len(number) != 10:
        raise InvalidLength()
    if calc_check_digit(number) != number[3]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid VAT number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



# pin.py - functions for handling Kenya PIN numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""PIN (Personal Identification Number, Kenya tax number).

The Personal Identification Number (KRA PIN) is an 11 digit unique number that
is issued by Kenya Revenue Authority (KRA) for purposes of transacting business
with KRA, other Government agencies and service providers. It can be issued for
individuals and non-individuals like companies, schools, organisations, etc.

The number consists of 11 characters, where the first one is an A (for
individuals) or a P (for non-individuals), the last one is a letter, and the
rest are digits.

More information:

* https://www.kra.go.ke/individual/individual-pin-registration/learn-about-pin/about-pin
* https://itax.kra.go.ke/KRA-Portal/pinChecker.htm

>>> validate('P051365947M')
'P051365947M'
>>> validate('A004416331M')
'A004416331M'
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('V1234567890')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> format('a004416331m')
'A004416331M'
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


# The number consists of 11 characters, where the first one is an A (for
# individuals) or a P (for non-individuals), the last one is a letter, and the
# rest are digits.
_pin_re = re.compile(r'^[A|P]{1}[0-9]{9}[A-Z]{1}$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -').strip().upper()


def validate(number):
    """Check if the number is a valid Kenya PIN number.

    This checks the length and formatting.
    """
    number = compact(number)
    if len(number) != 11:
        raise InvalidLength()
    match = _pin_re.search(number)
    if not match:
        raise InvalidFormat()
    return number


def is_valid(number):
    """Check if the number is a valid Kenya PIN number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Final

from opentelemetry.metrics import Histogram, Meter

DNS_LOOKUP_DURATION: Final = "dns.lookup.duration"
"""
Measures the time taken to perform a DNS lookup
Instrument: histogram
Unit: s
"""


def create_dns_lookup_duration(meter: Meter) -> Histogram:
    """Measures the time taken to perform a DNS lookup"""
    return meter.create_histogram(
        name=DNS_LOOKUP_DURATION,
        description="Measures the time taken to perform a DNS lookup.",
        unit="s",
    )



from typing import (
    Dict,
    Any,
)

from pos_model import OrderTender

from models import (
    NfeTender,
    MaxNfceFields,
)
from utils import get_t_pag


class TPagService(object):

    def __init__(
        self,            # type: TPagService
        tenders_type,    # type: Dict[str, Any]
    ):
        # type: (...) -> None

        self.tenders_type = tenders_type
        self.static_pix_qty = 0
        self.dynamic_pix_qty = 0

    def get_t_pag(
        self,      # type: TPagService
        tender,    # type: OrderTender
    ):
        # type: (...) -> int

        t_pag = get_t_pag(
            tender=tender,
            tenders_type=self.tenders_type,
        )
        if t_pag in [NfeTender.pix_estatico, NfeTender.pagamento_instantaneo_pix]:
            return self._get_new_t_pag_if_maximum_reached(t_pag=t_pag)

        return t_pag

    def _get_new_t_pag_if_maximum_reached(
        self,    # type: TPagService
        t_pag,   # type: int
    ):
        # type: (...) -> int

        if t_pag == NfeTender.pix_estatico and not self.static_pix_qty:
            self.static_pix_qty = 1
        elif t_pag == NfeTender.pagamento_instantaneo_pix:
            self.dynamic_pix_qty += 1

        if self._sum() > MaxNfceFields.max_obs_cont:
            return NfeTender.outros

        return t_pag

    def _sum(
        self,    # type: TPagService
    ):
        # type: (...) -> int

        return self.static_pix_qty + (self.dynamic_pix_qty * 2)



# iso6346.py - functions for handling ISO 6346
#
# Copyright (C) 2014 Openlabs Technologies & Consulting (P) Limited
# Copyright (C) 2014-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ISO 6346 (International standard for container identification)

ISO 6346 is an international standard covering the coding, identification and
marking of intermodal (shipping) containers used within containerized
intermodal freight transport. The standard establishes a visual identification
system for every container that includes a unique serial number (with check
digit), the owner, a country code, a size, type and equipment category as well
as any operational marks. The standard is managed by the International
Container Bureau (BIC).

More information:

* https://en.wikipedia.org/wiki/ISO_6346

>>> validate('csqu3054383')
'CSQU3054383'
>>> validate('CSQU3054384')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('tasu117 000 0')
'TASU 117000 0'
"""

import re

from stdnum.exceptions import *
from stdnum.util import clean


_iso6346_re = re.compile(r'^\w{3}(U|J|Z|R)\d{7}$')


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


def calc_check_digit(number):
    """Calculate check digit and return it for the 10 digit owner code and
    serial number."""
    number = compact(number)
    alphabet = '0123456789A BCDEFGHIJK LMNOPQRSTU VWXYZ'
    return str(sum(
        alphabet.index(n) * pow(2, i)
        for i, n in enumerate(number)) % 11 % 10)


def validate(number):
    """Validate the given number (unicode) for conformity to ISO 6346."""
    number = compact(number)
    if len(number) != 11:
        raise InvalidLength()
    if not _iso6346_re.match(number):
        raise InvalidFormat()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check whether the number conforms to the standard ISO6346. Unlike
    the validate function, this will not raise ValidationError(s)."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[:4], number[4:-1], number[-1:]))



class CannotSaveEmptyOrder(Exception):

    def __init__(
        self,       # type: CannotSaveEmptyOrder
    ):
        # type: (...) -> None
        super(CannotSaveEmptyOrder, self).__init__()

    @staticmethod
    def get_error_code():
        # type: (...) -> str

        return "CannotSaveEmptyOrder"

    def __str__(
        self,  # type: CannotSaveEmptyOrder
    ):
        # type: (...) -> str

        return self.get_error_code()



class BossException(Exception):

    inner_exceptions = {}

    def __init__(
        self,           # type: BossException
        error_code,     # type: int
        error_message,  # type: str
    ):
        # type: (...) -> None

        self.error_code = error_code
        self.error_message = error_message
        super(BossException, self).__init__("{} - {}".format(error_code, error_message))

    @classmethod
    def get_exception(
        cls,
        error_code,  # type: int
    ):
        # type: (...) -> BossException

        cls.get_inner_exceptions()
        exception_class = cls.inner_exceptions.get(error_code, UnexpectedError)
        return exception_class()

    @classmethod
    def get_inner_exceptions(
        cls,
    ):
        # type: (...) -> None

        for subclass in cls.__subclasses__():
            if hasattr(subclass, "error_code") and hasattr(subclass, "error_message"):
                cls.inner_exceptions[subclass.error_code] = subclass


class UnexpectedError(BossException):

    error_code = -1
    error_message = "Unexpected error"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class NoData(BossException):

    error_code = -2
    error_message = "No data"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class NoUtf8Data(BossException):

    error_code = -3
    error_message = "User data was not utf-8"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class InvalidDataFormat(BossException):

    error_code = -4
    error_message = "User data was not utf-8"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class NoUser(BossException):

    error_code = -5
    error_message = "User not found"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class NoPassword(BossException):

    error_code = -6
    error_message = "Invalid password data"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class AuthenticationFailed(BossException):

    error_code = -7
    error_message = "Authentication failed"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class AccessDenied(BossException):

    error_code = -8
    error_message = "Access denied"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )


class MissingMandatoryField(BossException):

    error_code = -9
    error_message = "Missing mandatory field"

    def __init__(
        self,  # type: BossException
    ):
        # type: (...) -> None

        BossException.__init__(
            self,
            error_code=self.error_code,
            error_message=self.error_message,
        )



# tin.py - functions for handling Austrian tax identification numbers
# coding: utf-8
#
# Copyright (C) 2017 Holvi Payment Services Oy
# Copyright (C) 2018 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

r"""Abgabenkontonummer (Austrian tax identification number).

The Austrian tax identification number (Abgabenkontonummer) consists of 2
digits for the issuing tax office (Finanzamtsnummer) and 7 digits for the
subject and a check digit (Steuernummer).

More information:

* https://de.wikipedia.org/wiki/Abgabenkontonummer
* https://service.bmf.gv.at/Service/Anwend/Behoerden/show_mast.asp

>>> validate('59-119/9013')
'591199013'
>>> validate('59-119/9013', office='St. Veit Wolfsberg')
'591199013'
>>> import json
>>> print(json.dumps(info('59-119/9013'), indent=2, sort_keys=True))
{
  "office": "St. Veit Wolfsberg",
  "region": "K\u00e4rnten"
}
>>> format('591199013')
'59-119/9013'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' -./,').strip()


def _min_fa(office):
    """Convert the tax office name to something that we can use for
    comparison without running into encoding issues."""
    return ''.join(
        x for x in office.lower()
        if x in 'bcdefghijklmnopqrstvwxyz')


def calc_check_digit(number):
    """Calculate the check digit."""
    number = compact(number)
    s = sum(
        int('0246813579'[int(n)]) if i % 2 else int(n)
        for i, n in enumerate(number[:8]))
    return str((10 - s) % 10)


def info(number):
    """Return a dictionary of data about the supplied number. This typically
    returns the the tax office and region."""
    number = compact(number)
    from stdnum import numdb
    return numdb.get('at/fa').info(number[:2])[0][1]


def validate(number, office=None):
    """Check if the number is a valid tax identification number. This checks
    the length and formatting. The tax office can be supplied to check that
    the number was issued in the specified tax office."""
    number = compact(number)
    if len(number) != 9:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    i = info(number)
    if not i:
        raise InvalidComponent()
    if office and _min_fa(i.get('office')) != _min_fa(office):
        raise InvalidComponent()
    return number


def is_valid(number, office=None):
    """Check if the number is a valid tax identification number. This checks
    the length, formatting and check digit."""
    try:
        return bool(validate(number, office))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '%s-%s/%s' % (number[:2], number[2:5], number[5:])



"""
Forward-ports of types from Python 2 for use with Python 3:

- ``basestring``: equivalent to ``(str, bytes)`` in ``isinstance`` checks
- ``dict``: with list-producing .keys() etc. methods
- ``str``: bytes-like, but iterating over them doesn't product integers
- ``long``: alias of Py3 int with ``L`` suffix in the ``repr``
- ``unicode``: alias of Py3 str with ``u`` prefix in the ``repr``

"""

from past import utils

if utils.PY2:
    import __builtin__
    basestring = __builtin__.basestring
    dict = __builtin__.dict
    str = __builtin__.str
    long = __builtin__.long
    unicode = __builtin__.unicode
    __all__ = []
else:
    from .basestring import basestring
    from .olddict import olddict
    from .oldstr import oldstr
    long = int
    unicode = str
    # from .unicode import unicode
    __all__ = ['basestring', 'olddict', 'oldstr', 'long', 'unicode']




from typing import Optional


class LoyaltyDiscountData(object):
    def __init__(
        self,                       # type: LoyaltyDiscountData
        new_payment_amount=None,    # type: Optional[float]
        discount_amount=None,       # type: Optional[float]
        benefit_id=None,            # type: Optional[str]
        loyalty_id=None,            # type: Optional[str]
        benefit_key=None,           # type: Optional[str]
    ):
        # type: (...) -> None

        self.new_payment_amount = new_payment_amount
        self.discount_amount = discount_amount
        self.benefit_id = benefit_id
        self.loyalty_id = loyalty_id
        self.benefit_key = benefit_key



from ._CustomProperty import CustomProperty
from ._DeliveryInfoDto import DeliveryInfoDto
from ._Order import Order
from ._OrderState import OrderState
from ._SaleLine import SaleLine
from ._SaleLineIdentifier import SaleLineIdentifier
from ._StateHistory import StateHistory
from ._Tender import Tender
from ._OrderState import OrderState
from .order_xml_parser import parse_order



import cfgtools
import logging
import os

from datetime import datetime
from typing import Optional, Tuple

from mw_helper import show_message_dialog

from actions.fiscal import get_nf_type
from application.interactor import FiscalWrapperInteractor


NFCE_TYPE = "NFCE"
IS_NFCE = None

alert_certificate_expiration = True
certificate_expiration_days = None      # type: Optional[int]
last_updated_date = None                # type: Optional[datetime]
first_alert_expiration_days = None      # type: Optional[int]
second_alert_expiration_days = None     # type: Optional[int]

logger = logging.getLogger("PosActions")


def check_certificate(
        pos_id,             # type: int
):
    # type: (...) -> None

    global alert_certificate_expiration

    if not _check_is_nfce():
        return

    now = datetime.now()
    if not last_updated_date or (now - last_updated_date).days >= 1:
        alert_certificate_expiration = True
        _set_certificate_expiration_days(now)

    first_alert_days, second_alert_days = _get_certificate_alert_days()

    ignore_alert = certificate_expiration_days > first_alert_days
    if ignore_alert:
        return

    must_show_first_alert = certificate_expiration_days >= second_alert_days and alert_certificate_expiration
    must_show_second_alert = certificate_expiration_days < second_alert_days
    if must_show_first_alert or must_show_second_alert:
        alert_certificate_expiration = False
        show_message_dialog(pos_id, message="$ALERT_CERTIFICATE_EXPIRATION_DAYS|{}".format(certificate_expiration_days))


def _check_is_nfce():
    # type: (...) -> bool

    global IS_NFCE

    if IS_NFCE is None:
        IS_NFCE = get_nf_type() == NFCE_TYPE

    return IS_NFCE


def _set_certificate_expiration_days(
    updated_time,   # type: datetime
):
    # type: (...) -> None

    global certificate_expiration_days
    global last_updated_date

    logger.info("Getting certificate expiration time")

    last_updated_date = updated_time.replace(hour=0, minute=0, second=0, microsecond=0)
    fiscal_wrapper_interactor = FiscalWrapperInteractor()
    try:
        certificate_expiration_days = fiscal_wrapper_interactor.get_certificate_expiration_days()
    except Exception:
        logger.exception("Error getting certificate expiration time")


def _get_certificate_alert_days():
    # type: (...) -> Tuple[int, int]

    global first_alert_expiration_days, second_alert_expiration_days

    if first_alert_expiration_days is None or second_alert_expiration_days is None:
        config = cfgtools.read(os.environ["LOADERCFG"])
        config_path = "Customizations.CertificateExpiration"
        first_alert_expiration_days = int(config.find_value(config_path + ".FirstAlertDays") or 15)
        second_alert_expiration_days = int(config.find_value(config_path + ".SecondAlertDays") or 3)

    return first_alert_expiration_days, second_alert_expiration_days



from typing import (
    Any,
    Dict,
)

from model import FiscalType
from .._DataBuilder import DataBuilder
from .._Util import build_util_variables
from nreports.builders.fiscal_coupon import (
    FiscalCouponNfceBuilder,
    FiscalCouponSatBuilder,
)
from nreports.builders.end_of_sale.fiscal_coupon_mobile import (
    BaseFiscalCouponMobileNfce,
    BaseFiscalCouponMobileSat,
)
from nreports.utils import get_fiscal_type
from nreports_api.shared import JSON


class CanceledFiscalCouponBuilder(DataBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        order_id = int(kwargs["orderId"])
        order_service = self.order_api_repository.get_order_service(pos_id=pos_id)
        kwargs["orderPicture"] = order_xml = order_service.get_order_picture(order_id=order_id)
        kwargs["fiscalXml"] = fiscal_xml = self.order_api_repository.get_fiscal_xml(
            pos_id=pos_id,
            order_id=order_id,
            order_xml=order_xml,
        )
        data_to_build = build_util_variables(
            constants=self.constants,
            data_to_build=data_to_build,
            **kwargs,
        )

        is_mobile = str(kwargs.get("isMobile", "false")).lower() == "true"
        fiscal_type = get_fiscal_type(fiscal_xml=fiscal_xml)
        if fiscal_type == FiscalType.NFCE:
            builder = BaseFiscalCouponMobileNfce if is_mobile else FiscalCouponNfceBuilder
            data_to_build["NFCE_CANCELED_COUPON"] = True
            data_to_build["IS_NFCE"] = fiscal_type == FiscalType.NFCE
        else:
            builder = BaseFiscalCouponMobileSat if is_mobile else FiscalCouponSatBuilder
            data_to_build["SAT_CANCELED_COUPON"] = True
            data_to_build["IS_SAT"] = fiscal_type == FiscalType.SAT

        data_to_build = builder(*self.get_builder_init()).build_data(
            data_to_build=data_to_build,
            **kwargs,
        )

        return data_to_build



import cPickle as Pickle
from datetime import datetime

import pytz
from typing import List, Optional
from tzlocal import get_localzone
from ._CustomStorage import CustomStorage
from ._Item import Item
from ._ProdStates import ProdStates
from ._StateEvent import StateEvent
from ._TagEvent import TagEvent

time_offset = datetime.now() - datetime.utcnow()


class ProductionOrder(CustomStorage):

    """ class ProductionOrder(CustomStorage)
    Represents a production order.
    Instances of this class can be handled and/or modified by any production-box.
    Use the methods write_data, read_data and delete_data in order to modify instances
    in a safe way (avoiding name conflicts).
    """
    # Note - NEVER use mutable objects as the default value here (they are shared by all instances)
    # Mutable objects MUST be created in the constructor
    order_id = 0  #: Order identification number
    state = ""  #: Current order state ("IN_PROGRESS", "TOTALED", ..etc..)
    state_id = 0  #: Current state number
    prod_sequence = ""  #: Production sequence number used to sort orders - SHOULD ONLY BE SET ON A CLONED ORDER
    pos_id = 0  #: Current order POS id
    originator = ""  #: Name of the order originator
    pod_type = ""  #: The point-of-distribution of this order ("FC", "DT")
    created_at = ""  #: Local timestamp of the "order start" (in database format: YYYY-MM-DDTHH:MM:SS[.SSS])
    business_day = ""
    created_at_gmt = ""  #: UTC timestamp of the "order start" (in database format: YYYY-MM-DDTHH:MM:SS[.SSS])
    operator_id = 0  #: ID of the operator who took this order
    operator_name = ""
    operator_level = 0
    order_type = ""  #: Order type ("SALE", "REFUND", ..etc..)
    order_subtype = ""  #: Order subtype (script-defined, and generally not known by the kernel)
    sale_type = ""  #: Current sale type (EAT_IN, TAKE_OUT, etc...)
    total_gross = ""  #: Order total gross amount
    product_priority = 0  #: Product display priority
    state_history = None  #: List of StateEvent objects of this order
    items = None  #: Level-0 items of this order
    properties = None  #: List of OrderProperties from the Order Picture
    cloned = False  #: Indicates if this is a cloned instance of an order
    purged = False  #: Indicates if this order has just been purged from the database
    session_id = ""  #: Session ID for this order
    display_time = ""  #: Time to be displayed in the KDS
    tagged_timestamp = None  #: Time when the whole order was tagged
    recalled = False  #: Whether this order was called back after being served
    first_processing = False  #: Indicates if this is the first processing time
    max_prep_time = None
    cost = None
    skip_course = False
    round_robin_path = ""
    tags = None
    current_course = ''
    jit_printer_name = ''
    order_id_count = ''
    _prod_state_history = None
    selected_pickup_area = None
    stored_split_order = False

    def __init__(self, **kargs):
        super(ProductionOrder, self).__init__()
        self.state_history = []  # type: List[StateEvent]
        self.items = []  # type: List[Item]
        self.properties = {}
        self.undo_list = []  # type: List[ChangeProdStateUndoCommand]
        self.tag_history = []  # type: List[TagEvent]
        self.cost = None
        self._prod_state_history = []  # type: List[StateEvent]
        self.tags = {}
        self.selected_pickup_area = {}
        self.showed_lines = {}
        for key, val in kargs.items():
            setattr(self, key, val)

    @property
    def prod_state(self):
        if self._prod_state_history:
            return self._prod_state_history[-1].prod_state
        return ProdStates.NORMAL

    @prod_state.setter
    def prod_state(self, new_state):
        if self._prod_state_history and new_state == self._prod_state_history[-1].prod_state:
            return

        date_now = datetime.now()
        timezone = pytz.timezone(get_localzone().zone)
        date_now_tz = timezone.localize(date_now).strftime("%Y-%m-%dT%H:%M:%S%z")
        self._prod_state_history.append(StateEvent(prod_state=new_state, timestamp=date_now_tz))

    @property
    def prod_state_last_update(self):
        if self._prod_state_history:
            return self._prod_state_history[-1].timestamp
        return None

    @property
    def prod_state_history(self):
        return tuple(self._prod_state_history or [])

    @property
    def display_time_gmt(self):
        try:
            timestamp_date = datetime.strptime(self.display_time[0:19], "%Y-%m-%dT%H:%M:%S")
            gmt_date = timestamp_date - time_offset
            return gmt_date.isoformat() + "Z"
        except ValueError:
            return ""

    def __str__(self):
        ret = ""
        for item in self.items:
            ret += str(item) + "\n"

        return "Id: {}, ProdState: {}, Sequence: {}, SaleType: {}, State: {}{}{}{}\nItems:\n{}".format(
            self.order_id,
            self.prod_state,
            self.prod_sequence,
            self.sale_type,
            self.state,
            ", MaxPrepTime: {}".format(self.max_prep_time) if self.max_prep_time else "",
            ", Cost: {}".format(self.cost) if self.cost is not None else "",
            ", RoundRobinPath: {}".format(self.round_robin_path) if hasattr(self, "round_robin_path") else "",
            ret
        )

    def __repr__(self):
        return "(%s) %s" % (object.__repr__(self), str(self))

    def __eq__(self, other):
        # type: (ProductionOrder) -> bool
        prop_list_to_compare = [
            "order_id",
            "state_history",
            "items",
            "prod_state",
            "tags",
            "display_time",
            "properties",
            "selected_pickup_area",
        ]

        def _dump_prop(order, prop):
            return Pickle.dumps(getattr(order, prop), Pickle.HIGHEST_PROTOCOL)

        if not other:
            return False

        if not isinstance(other, ProductionOrder):
            return NotImplemented

        return not any(_dump_prop(other, p) != _dump_prop(self, p) for p in prop_list_to_compare)

    def add_tag(self, tag):
        if tag not in self.tags:
            date_now = datetime.now()
            timezone = pytz.timezone(get_localzone().zone)
            date_now_tz = timezone.localize(date_now).strftime("%Y-%m-%dT%H:%M:%S%z")

            self.tags[tag] = date_now_tz

    def remove_tag(self, tag):
        if tag in self.tags:
            del self.tags[tag]

    def add_line_tag(self, line_id, tag, line_jit_list=None):
        item = self.get_item_by_line_id(line_id)
        if item is None:
            return

        item.add_tag(tag, line_jit_list)
        if item.is_option():
            self._add_line_sons_tags(item, tag, line_jit_list)

        self.build_showed_lines()

    def _add_line_sons_tags(self, item, tag, line_jit_list=None):
        for son in item.items:
            son.add_tag(tag, line_jit_list)
            if son.items:
                self._add_line_sons_tags(son, tag, line_jit_list)

    def remove_line_tag(self, line_id, tag, line_jit_list=None):
        item = self.get_item_by_line_id(line_id)
        if item is None:
            return

        item.remove_tag(tag, line_jit_list)
        if item.is_option():
            self._remove_line_sons_tags(item, tag, line_jit_list)

        self.build_showed_lines()

    def _remove_line_sons_tags(self, item, tag, line_jit_list=None):
        for son in item.items:
            son.remove_tag(tag, line_jit_list)
            if son.items:
                self._remove_line_sons_tags(son, tag, line_jit_list)

    def toggle_line_tag(self, line_id, tag, line_jit_list=None):
        item = self.get_item_by_line_id(line_id)
        if item is None:
            return

        jit_tags = item.toggle_tag(tag, line_jit_list)
        for jit_tag in jit_tags:
            if item.has_tag(jit_tag):
                self._add_line_sons_tags(item, tag, line_jit_list)
            else:
                self._remove_line_sons_tags(item, tag, line_jit_list)

    def build_showed_lines(self):
        self.showed_lines = {}
        count = 1
        for item in self.items:
            count = self.build_item_tags(item, count)

    def build_item_tags(self, item, count):
        if item.is_ingredient():
            return count

        has_show = False
        for tag in item.get_tags():
            if tag.startswith("show"):
                has_show = True
                break

        line_id = item.get_line_id()
        if has_show:
            if line_id not in self.showed_lines:
                if item.measure_unit == "UN":
                    qty = int(item.multiplied_qty)
                else:
                    qty = 1

                self.showed_lines[line_id] = list(range(count, count + qty))
                count += qty
        else:
            if line_id in self.showed_lines:
                del self.showed_lines[line_id]

        for son in item.items:
            count = self.build_item_tags(son, count)

        return count

    def get_item_by_line_id(
        self,           # type: ProductionOrder
        line_id,        # type: str
    ):
        # type: (...) -> Optional[Item]

        for item in self.items:
            found = self.get_item_or_son(line_id, item)
            if found is not None:
                return found

            for joined_item in item.joined_items:
                found = self.get_item_or_son(line_id, joined_item)
                if found is not None:
                    return found

        return None

    def get_item_or_son(self, line_id, item):
        if item.get_line_id() == line_id:
            return item

        for son in item.items:
            found = self.get_item_or_son(line_id, son)
            if found is not None:
                return found

        return None

    def get_item(self, item):
        # type: (Item) -> Optional[Item]

        for current_item in self.items:
            if current_item.line_number == item.line_number and current_item.item_code == item.item_code:
                return current_item
        return None

    def clone(self, items=None, index=None):
        # type: (Optional[List[Item]], Optional[int]) -> ProductionOrder

        cloned_order = Pickle.loads(Pickle.dumps(self, -1))
        if items is not None:
            cloned_order.items = Pickle.loads(Pickle.dumps(items, -1))

        cloned_order.order_id = str(self.order_id) + "-" + str(index) if index is not None else self.order_id
        return cloned_order

    def get_cloned_items(
        self,               # type: ProductionOrder
        allowed_items,      # type: List[Item]
    ):
        # type: (...) -> List[Item]

        cloned_items = Pickle.loads(Pickle.dumps(self.items, -1))
        items_ids = [x.get_line_id() for x in allowed_items]

        items = self._get_items(
            items=cloned_items,
            allowed_items_id=items_ids,
            include_all_sons_from_allowed_items=True,
        )

        return items

    def _get_items(
        self,                                   # type: ProductionOrder
        items,                                  # type: List[Item]
        allowed_items_id,                       # type: List[str]
        include_all_sons_from_allowed_items,    # type: bool
    ):
        if not items:
            return []

        new_items = []
        for item in items:
            item_allowed = item.get_line_id() in allowed_items_id
            if item_allowed and include_all_sons_from_allowed_items:
                sons_items = item.items
            else:
                sons_items = self._get_items(
                    items=item.items,
                    allowed_items_id=allowed_items_id,
                    include_all_sons_from_allowed_items=include_all_sons_from_allowed_items,
                )

            if item_allowed or sons_items:
                item.items = sons_items
                new_items.append(item)

        return new_items

    def equal_items(self, order):
        if order is None:
            return False

        if isinstance(order, ProductionOrder) is False:
            return False

        for item in order.items:
            saved_item = self.get_item(item)
            if saved_item is None:
                return False
            if compare_items(item, saved_item, is_level0=True) is False:
                return False

        for item in self.items:
            other_item = order.get_item(item)
            if other_item is None:
                return False
            if compare_items(item, other_item, is_level0=True) is False:
                return False
        return True


def compare_items(item1, item2, is_level0=False):
    if item1 is None or item2 is None:
        return False

    items1, items2 = real_items(item1.items), real_items(item2.items)
    if (item1.part_code == item2.part_code) and (len(items1) == len(items2)) and (item1.level == item2.level):
        if (not is_level0) and item1.qty != item2.qty:
            return False  # Only consider the quantity for non-level0
        if compare_comments(item1, item2) is False:
            # Items have different comments
            return False
        if items1 and not all(map(compare_items, items1, items2)):
            return False
        return True


def compare_comments(item1, item2):
    if item1 is None or item2 is None:
        return False

    none_comments = bool(item1.comments is None) + bool(item2.comments is None)
    if none_comments == 0:
        return True

    if none_comments == 1:
        return False

    if item1.comments.keys() != item2.comments.keys():
        return False

    for key in item1.comments:
        item1_comment = item1.comments[key]
        item2_comment = item2.comments[key]
        if item1_comment.comment_id != item2_comment.comment_id:
            return False
        if item1_comment.text != item2_comment.text:
            return False

    return True


def real_items(item_list):
    def _real_item(item):
        type_ok = (item.item_type not in ("INGREDIENT", "CANADD"))
        return type_ok or (item.default_qty != item.qty) or (item.only == "true")

    return [x for x in item_list if _real_item(x)]



from ._AckMessage import AckMessage
from ._ApplicationErrorMessage import ApplicationErrorMessage
from ._BadRequestMessage import BadRequestMessage
from ._DataType import DataType
from ._DefaultToken import DefaultToken
from ._Event import Event
from ._HttpRequestMessage import HttpRequestMessage
from ._HttpResponseMessage import HttpResponseMessage
from ._InternalErrorMessage import InternalErrorMessage
from ._JsonHttpResponseMessage import JsonHttpResponseMessage
from ._JsonSwaggerMessage import JsonSwaggerMessage
from ._Message import Message
from ._MessageBus import MessageBus
from ._MessageBusException import MessageBusException
from ._MessageBusExceptionType import MessageBusExceptionType
from ._NakMessage import NakMessage
from ._SubscriptionInfo import SubscriptionInfo
from ._SwaggerMessage import SwaggerMessage
from ._Timeout import Timeout
from ._TokenCreator import TokenCreator
from ._TokenPriority import TokenPriority



import json

from xml.etree import cElementTree as eTree

from msgbus import MBEasyContext, TK_SYS_ACK
from pos_model import TaxItem, SaleItem
from typing import Dict, List


class CalculateTaxesItemCommon(object):
    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        self.mb_context = mb_context

    def get_taxes_item(
        self,               # type: CalculateTaxesItemCommon
        sale_line,          # type: SaleItem
        taxes_calc,         # type: str
        base_reduction,     # type: Dict
        cst,                # type: Dict
    ):
        # type: (...) -> List[TaxItem]

        data = json.dumps(dict(
            partCode=sale_line.part_code,
            productName=sale_line.product_name,
            quantity=sale_line.quantity,
            unitPrice=sale_line.unit_price,
            itemPrice=sale_line.item_price,
            taxesCalc=taxes_calc,
            baseReduction=base_reduction,
            cst=cst,
        )).encode('utf-8')

        msg = self.mb_context.MB_EasyEvtSend("TAX_CALC_ITEM", "", data, synchronous=True)
        if msg.token == TK_SYS_ACK:
            return self._tax_decode(data=data)

        return []

    @staticmethod
    def _tax_decode(
        data,   # type: str
    ):
        # type: (...) -> List[TaxItem]

        tax_items_xml = eTree.XML(data)
        tax_items = []

        for tax_item in tax_items_xml.findall("TaxItem"):
            tax = TaxItem()
            tax.tax_name = tax_item.get("taxName")
            tax.tax_rate = tax_item.get("taxRate")
            tax.tax_index = tax_item.get("taxIndex")
            tax.tax_rule_id = tax_item.get("taxRuleId")
            tax.base_amount_ad = tax_item.get("baseAmountAD")
            tax.base_amount_bd = tax_item.get("baseAmountBD")
            tax.base_reduced_amount = tax_item.get("baseReducedAmount")
            tax.tax_amount_ad = tax_item.get("taxAmountAD")
            tax.tax_amount_bd = tax_item.get("taxAmountBD")
            tax.tax_included = tax_item.get("taxIncluded")

            tax_items.append(tax)

        return tax_items



from logging import Logger
from typing import Dict

from messagebus import MessageBus

from production.manager import ProductionManager
from production.model import PrinterNameFoundException
from production.treebuilder import ViewConfiguration
from production.view import (
    PrinterView,
    PicklistPrinterView,
    ReportsPrinterView,
)
from production.view.printerviewservices.reportgenerator import PickListReportGenerator
from production.view.printerviewservices.reportprinter import PickListPrinter

from ._CommonViewBuilder import (
    get_extra,
    get_logger,
)


class PrinterViewBuilder(object):
    def __init__(
        self,                           # type: PrinterViewBuilder
        message_bus,                    # type: MessageBus
        loggers,                        # type: Dict[str, Logger]
        production_manager,             # type: ProductionManager
        timeout_to_cleanup_devices,     # type: int
    ):
        # type: (...) -> None

        self.production_manager = production_manager
        self.message_bus = message_bus
        self.loggers = loggers
        self.timeout_to_cleanup_devices = timeout_to_cleanup_devices

    def build(self, config):
        # type: (ViewConfiguration) -> PrinterView

        logger = get_logger(config, self.loggers)

        if "PrinterName" not in config.extra_config:
            raise PrinterNameFoundException("PrinterView without printer: {}".format(config.name))

        if config.extra_config.get("ReportType") == "nreports":
            return ReportsPrinterView(
                name=config.name,
                report_name=config.extra_config["ReportName"],
                printer_name=config.extra_config["PrinterName"],
                pick_list_name=config.extra_config.get("PickListName", ""),
                production_manager=self.production_manager,
                timeout=int(get_extra(config, "Timeout", 10000)) * 1000,
                message_bus=self.message_bus,
                tags_on_print=get_extra(config, "TagOnPrint", "printed").split(";"),
                logger=logger,
            )

        return self.build_picklist_printer_view(
            config=config,
            logger=logger,
        )

    def build_picklist_printer_view(
        self,       # type: PrinterViewBuilder
        config,     # type: ViewConfiguration
        logger,     # type: Logger
    ):
        # type: (...) -> PicklistPrinterView

        picklist_report_generator = PickListReportGenerator(
            message_bus=self.message_bus,
            configuration_json_path=config.extra_config.get("ConfigurationJsonPath", "").encode("utf-8"),
            picklist_name=config.extra_config.get("PickListName", "").encode("utf-8"),
            logger=logger
        )

        picklist_printer = PickListPrinter(
            message_bus=self.message_bus,
            timeout=int(get_extra(config, "Timeout", 10000)) * 1000,
            name=config.name,
            logger=logger,
            kds_id=str(config.extra_config.get("KdsId", "").replace("KDS", "") or ""),
            printer=config.extra_config["PrinterName"].encode("utf-8"),
            timeout_to_cleanup_devices=self.timeout_to_cleanup_devices,
        )

        return PicklistPrinterView(
            name=config.name,
            production_manager=self.production_manager,
            report_generator=picklist_report_generator,
            picklist_printer=picklist_printer,
            tags_on_print=get_extra(config, "TagOnPrint", "printed").split(";"),
            logger=get_logger(config, self.loggers),
        )



# orm/mapped_collection.py
# Copyright (C) 2005-2024 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php

from __future__ import annotations

import operator
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import TypeVar
from typing import Union

from . import base
from .collections import collection
from .collections import collection_adapter
from .. import exc as sa_exc
from .. import util
from ..sql import coercions
from ..sql import expression
from ..sql import roles
from ..util.langhelpers import Missing
from ..util.langhelpers import MissingOr
from ..util.typing import Literal

if TYPE_CHECKING:
    from . import AttributeEventToken
    from . import Mapper
    from .collections import CollectionAdapter
    from ..sql.elements import ColumnElement

_KT = TypeVar("_KT", bound=Any)
_VT = TypeVar("_VT", bound=Any)


class _PlainColumnGetter(Generic[_KT]):
    """Plain column getter, stores collection of Column objects
    directly.

    Serializes to a :class:`._SerializableColumnGetterV2`
    which has more expensive __call__() performance
    and some rare caveats.

    """

    __slots__ = ("cols", "composite")

    def __init__(self, cols: Sequence[ColumnElement[_KT]]) -> None:
        self.cols = cols
        self.composite = len(cols) > 1

    def __reduce__(
        self,
    ) -> Tuple[
        Type[_SerializableColumnGetterV2[_KT]],
        Tuple[Sequence[Tuple[Optional[str], Optional[str]]]],
    ]:
        return _SerializableColumnGetterV2._reduce_from_cols(self.cols)

    def _cols(self, mapper: Mapper[_KT]) -> Sequence[ColumnElement[_KT]]:
        return self.cols

    def __call__(self, value: _KT) -> MissingOr[Union[_KT, Tuple[_KT, ...]]]:
        state = base.instance_state(value)
        m = base._state_mapper(state)

        key: List[_KT] = [
            m._get_state_attr_by_column(state, state.dict, col)
            for col in self._cols(m)
        ]
        if self.composite:
            return tuple(key)
        else:
            obj = key[0]
            if obj is None:
                return Missing
            else:
                return obj


class _SerializableColumnGetterV2(_PlainColumnGetter[_KT]):
    """Updated serializable getter which deals with
    multi-table mapped classes.

    Two extremely unusual cases are not supported.
    Mappings which have tables across multiple metadata
    objects, or which are mapped to non-Table selectables
    linked across inheriting mappers may fail to function
    here.

    """

    __slots__ = ("colkeys",)

    def __init__(
        self, colkeys: Sequence[Tuple[Optional[str], Optional[str]]]
    ) -> None:
        self.colkeys = colkeys
        self.composite = len(colkeys) > 1

    def __reduce__(
        self,
    ) -> Tuple[
        Type[_SerializableColumnGetterV2[_KT]],
        Tuple[Sequence[Tuple[Optional[str], Optional[str]]]],
    ]:
        return self.__class__, (self.colkeys,)

    @classmethod
    def _reduce_from_cols(cls, cols: Sequence[ColumnElement[_KT]]) -> Tuple[
        Type[_SerializableColumnGetterV2[_KT]],
        Tuple[Sequence[Tuple[Optional[str], Optional[str]]]],
    ]:
        def _table_key(c: ColumnElement[_KT]) -> Optional[str]:
            if not isinstance(c.table, expression.TableClause):
                return None
            else:
                return c.table.key  # type: ignore

        colkeys = [(c.key, _table_key(c)) for c in cols]
        return _SerializableColumnGetterV2, (colkeys,)

    def _cols(self, mapper: Mapper[_KT]) -> Sequence[ColumnElement[_KT]]:
        cols: List[ColumnElement[_KT]] = []
        metadata = getattr(mapper.local_table, "metadata", None)
        for ckey, tkey in self.colkeys:
            if tkey is None or metadata is None or tkey not in metadata:
                cols.append(mapper.local_table.c[ckey])  # type: ignore
            else:
                cols.append(metadata.tables[tkey].c[ckey])
        return cols


def column_keyed_dict(
    mapping_spec: Union[Type[_KT], Callable[[_KT], _VT]],
    *,
    ignore_unpopulated_attribute: bool = False,
) -> Type[KeyFuncDict[_KT, _KT]]:
    """A dictionary-based collection type with column-based keying.

    .. versionchanged:: 2.0 Renamed :data:`.column_mapped_collection` to
       :class:`.column_keyed_dict`.

    Returns a :class:`.KeyFuncDict` factory which will produce new
    dictionary keys based on the value of a particular :class:`.Column`-mapped
    attribute on ORM mapped instances to be added to the dictionary.

    .. note:: the value of the target attribute must be assigned with its
       value at the time that the object is being added to the
       dictionary collection.   Additionally, changes to the key attribute
       are **not tracked**, which means the key in the dictionary is not
       automatically synchronized with the key value on the target object
       itself.  See :ref:`key_collections_mutations` for further details.

    .. seealso::

        :ref:`orm_dictionary_collection` - background on use

    :param mapping_spec: a :class:`_schema.Column` object that is expected
     to be mapped by the target mapper to a particular attribute on the
     mapped class, the value of which on a particular instance is to be used
     as the key for a new dictionary entry for that instance.
    :param ignore_unpopulated_attribute:  if True, and the mapped attribute
     indicated by the given :class:`_schema.Column` target attribute
     on an object is not populated at all, the operation will be silently
     skipped.  By default, an error is raised.

     .. versionadded:: 2.0 an error is raised by default if the attribute
        being used for the dictionary key is determined that it was never
        populated with any value.  The
        :paramref:`_orm.column_keyed_dict.ignore_unpopulated_attribute`
        parameter may be set which will instead indicate that this condition
        should be ignored, and the append operation silently skipped.
        This is in contrast to the behavior of the 1.x series which would
        erroneously populate the value in the dictionary with an arbitrary key
        value of ``None``.


    """
    cols = [
        coercions.expect(roles.ColumnArgumentRole, q, argname="mapping_spec")
        for q in util.to_list(mapping_spec)
    ]
    keyfunc = _PlainColumnGetter(cols)
    return _mapped_collection_cls(
        keyfunc,
        ignore_unpopulated_attribute=ignore_unpopulated_attribute,
    )


class _AttrGetter:
    __slots__ = ("attr_name", "getter")

    def __init__(self, attr_name: str):
        self.attr_name = attr_name
        self.getter = operator.attrgetter(attr_name)

    def __call__(self, mapped_object: Any) -> Any:
        obj = self.getter(mapped_object)
        if obj is None:
            state = base.instance_state(mapped_object)
            mp = state.mapper
            if self.attr_name in mp.attrs:
                dict_ = state.dict
                obj = dict_.get(self.attr_name, base.NO_VALUE)
                if obj is None:
                    return Missing
            else:
                return Missing

        return obj

    def __reduce__(self) -> Tuple[Type[_AttrGetter], Tuple[str]]:
        return _AttrGetter, (self.attr_name,)


def attribute_keyed_dict(
    attr_name: str, *, ignore_unpopulated_attribute: bool = False
) -> Type[KeyFuncDict[Any, Any]]:
    """A dictionary-based collection type with attribute-based keying.

    .. versionchanged:: 2.0 Renamed :data:`.attribute_mapped_collection` to
       :func:`.attribute_keyed_dict`.

    Returns a :class:`.KeyFuncDict` factory which will produce new
    dictionary keys based on the value of a particular named attribute on
    ORM mapped instances to be added to the dictionary.

    .. note:: the value of the target attribute must be assigned with its
       value at the time that the object is being added to the
       dictionary collection.   Additionally, changes to the key attribute
       are **not tracked**, which means the key in the dictionary is not
       automatically synchronized with the key value on the target object
       itself.  See :ref:`key_collections_mutations` for further details.

    .. seealso::

        :ref:`orm_dictionary_collection` - background on use

    :param attr_name: string name of an ORM-mapped attribute
     on the mapped class, the value of which on a particular instance
     is to be used as the key for a new dictionary entry for that instance.
    :param ignore_unpopulated_attribute:  if True, and the target attribute
     on an object is not populated at all, the operation will be silently
     skipped.  By default, an error is raised.

     .. versionadded:: 2.0 an error is raised by default if the attribute
        being used for the dictionary key is determined that it was never
        populated with any value.  The
        :paramref:`_orm.attribute_keyed_dict.ignore_unpopulated_attribute`
        parameter may be set which will instead indicate that this condition
        should be ignored, and the append operation silently skipped.
        This is in contrast to the behavior of the 1.x series which would
        erroneously populate the value in the dictionary with an arbitrary key
        value of ``None``.


    """

    return _mapped_collection_cls(
        _AttrGetter(attr_name),
        ignore_unpopulated_attribute=ignore_unpopulated_attribute,
    )


def keyfunc_mapping(
    keyfunc: Callable[[Any], Any],
    *,
    ignore_unpopulated_attribute: bool = False,
) -> Type[KeyFuncDict[_KT, Any]]:
    """A dictionary-based collection type with arbitrary keying.

    .. versionchanged:: 2.0 Renamed :data:`.mapped_collection` to
       :func:`.keyfunc_mapping`.

    Returns a :class:`.KeyFuncDict` factory with a keying function
    generated from keyfunc, a callable that takes an entity and returns a
    key value.

    .. note:: the given keyfunc is called only once at the time that the
       target object is being added to the collection.   Changes to the
       effective value returned by the function are not tracked.


    .. seealso::

        :ref:`orm_dictionary_collection` - background on use

    :param keyfunc: a callable that will be passed the ORM-mapped instance
     which should then generate a new key to use in the dictionary.
     If the value returned is :attr:`.LoaderCallableStatus.NO_VALUE`, an error
     is raised.
    :param ignore_unpopulated_attribute:  if True, and the callable returns
     :attr:`.LoaderCallableStatus.NO_VALUE` for a particular instance, the
     operation will be silently skipped.  By default, an error is raised.

     .. versionadded:: 2.0 an error is raised by default if the callable
        being used for the dictionary key returns
        :attr:`.LoaderCallableStatus.NO_VALUE`, which in an ORM attribute
        context indicates an attribute that was never populated with any value.
        The :paramref:`_orm.mapped_collection.ignore_unpopulated_attribute`
        parameter may be set which will instead indicate that this condition
        should be ignored, and the append operation silently skipped. This is
        in contrast to the behavior of the 1.x series which would erroneously
        populate the value in the dictionary with an arbitrary key value of
        ``None``.


    """
    return _mapped_collection_cls(
        keyfunc, ignore_unpopulated_attribute=ignore_unpopulated_attribute
    )


class KeyFuncDict(Dict[_KT, _VT]):
    """Base for ORM mapped dictionary classes.

    Extends the ``dict`` type with additional methods needed by SQLAlchemy ORM
    collection classes. Use of :class:`_orm.KeyFuncDict` is most directly
    by using the :func:`.attribute_keyed_dict` or
    :func:`.column_keyed_dict` class factories.
    :class:`_orm.KeyFuncDict` may also serve as the base for user-defined
    custom dictionary classes.

    .. versionchanged:: 2.0 Renamed :class:`.MappedCollection` to
       :class:`.KeyFuncDict`.

    .. seealso::

        :func:`_orm.attribute_keyed_dict`

        :func:`_orm.column_keyed_dict`

        :ref:`orm_dictionary_collection`

        :ref:`orm_custom_collection`


    """

    def __init__(
        self,
        keyfunc: Callable[[Any], Any],
        *dict_args: Any,
        ignore_unpopulated_attribute: bool = False,
    ) -> None:
        """Create a new collection with keying provided by keyfunc.

        keyfunc may be any callable that takes an object and returns an object
        for use as a dictionary key.

        The keyfunc will be called every time the ORM needs to add a member by
        value-only (such as when loading instances from the database) or
        remove a member.  The usual cautions about dictionary keying apply-
        ``keyfunc(object)`` should return the same output for the life of the
        collection.  Keying based on mutable properties can result in
        unreachable instances "lost" in the collection.

        """
        self.keyfunc = keyfunc
        self.ignore_unpopulated_attribute = ignore_unpopulated_attribute
        super().__init__(*dict_args)

    @classmethod
    def _unreduce(
        cls,
        keyfunc: Callable[[Any], Any],
        values: Dict[_KT, _KT],
        adapter: Optional[CollectionAdapter] = None,
    ) -> "KeyFuncDict[_KT, _KT]":
        mp: KeyFuncDict[_KT, _KT] = KeyFuncDict(keyfunc)
        mp.update(values)
        # note that the adapter sets itself up onto this collection
        # when its `__setstate__` method is called
        return mp

    def __reduce__(
        self,
    ) -> Tuple[
        Callable[[_KT, _KT], KeyFuncDict[_KT, _KT]],
        Tuple[Any, Union[Dict[_KT, _KT], Dict[_KT, _KT]], CollectionAdapter],
    ]:
        return (
            KeyFuncDict._unreduce,
            (
                self.keyfunc,
                dict(self),
                collection_adapter(self),
            ),
        )

    @util.preload_module("sqlalchemy.orm.attributes")
    def _raise_for_unpopulated(
        self,
        value: _KT,
        initiator: Union[AttributeEventToken, Literal[None, False]] = None,
        *,
        warn_only: bool,
    ) -> None:
        mapper = base.instance_state(value).mapper

        attributes = util.preloaded.orm_attributes

        if not isinstance(initiator, attributes.AttributeEventToken):
            relationship = "unknown relationship"
        elif initiator.key in mapper.attrs:
            relationship = f"{mapper.attrs[initiator.key]}"
        else:
            relationship = initiator.key

        if warn_only:
            util.warn(
                f"Attribute keyed dictionary value for "
                f"attribute '{relationship}' was None; this will raise "
                "in a future release. "
                f"To skip this assignment entirely, "
                f'Set the "ignore_unpopulated_attribute=True" '
                f"parameter on the mapped collection factory."
            )
        else:
            raise sa_exc.InvalidRequestError(
                "In event triggered from population of "
                f"attribute '{relationship}' "
                "(potentially from a backref), "
                f"can't populate value in KeyFuncDict; "
                "dictionary key "
                f"derived from {base.instance_str(value)} is not "
                f"populated. Ensure appropriate state is set up on "
                f"the {base.instance_str(value)} object "
                f"before assigning to the {relationship} attribute. "
                f"To skip this assignment entirely, "
                f'Set the "ignore_unpopulated_attribute=True" '
                f"parameter on the mapped collection factory."
            )

    @collection.appender  # type: ignore[misc]
    @collection.internally_instrumented  # type: ignore[misc]
    def set(
        self,
        value: _KT,
        _sa_initiator: Union[AttributeEventToken, Literal[None, False]] = None,
    ) -> None:
        """Add an item by value, consulting the keyfunc for the key."""

        key = self.keyfunc(value)

        if key is base.NO_VALUE:
            if not self.ignore_unpopulated_attribute:
                self._raise_for_unpopulated(
                    value, _sa_initiator, warn_only=False
                )
            else:
                return
        elif key is Missing:
            if not self.ignore_unpopulated_attribute:
                self._raise_for_unpopulated(
                    value, _sa_initiator, warn_only=True
                )
                key = None
            else:
                return

        self.__setitem__(key, value, _sa_initiator)  # type: ignore[call-arg]

    @collection.remover  # type: ignore[misc]
    @collection.internally_instrumented  # type: ignore[misc]
    def remove(
        self,
        value: _KT,
        _sa_initiator: Union[AttributeEventToken, Literal[None, False]] = None,
    ) -> None:
        """Remove an item by value, consulting the keyfunc for the key."""

        key = self.keyfunc(value)

        if key is base.NO_VALUE:
            if not self.ignore_unpopulated_attribute:
                self._raise_for_unpopulated(
                    value, _sa_initiator, warn_only=False
                )
            return
        elif key is Missing:
            if not self.ignore_unpopulated_attribute:
                self._raise_for_unpopulated(
                    value, _sa_initiator, warn_only=True
                )
                key = None
            else:
                return

        # Let self[key] raise if key is not in this collection
        # testlib.pragma exempt:__ne__
        if self[key] != value:
            raise sa_exc.InvalidRequestError(
                "Can not remove '%s': collection holds '%s' for key '%s'. "
                "Possible cause: is the KeyFuncDict key function "
                "based on mutable properties or properties that only obtain "
                "values after flush?" % (value, self[key], key)
            )
        self.__delitem__(key, _sa_initiator)  # type: ignore[call-arg]


def _mapped_collection_cls(
    keyfunc: Callable[[Any], Any], ignore_unpopulated_attribute: bool
) -> Type[KeyFuncDict[_KT, _KT]]:
    class _MKeyfuncMapped(KeyFuncDict[_KT, _KT]):
        def __init__(self, *dict_args: Any) -> None:
            super().__init__(
                keyfunc,
                *dict_args,
                ignore_unpopulated_attribute=ignore_unpopulated_attribute,
            )

    return _MKeyfuncMapped


MappedCollection = KeyFuncDict
"""A synonym for :class:`.KeyFuncDict`.

.. versionchanged:: 2.0 Renamed :class:`.MappedCollection` to
   :class:`.KeyFuncDict`.

"""

mapped_collection = keyfunc_mapping
"""A synonym for :func:`_orm.keyfunc_mapping`.

.. versionchanged:: 2.0 Renamed :data:`.mapped_collection` to
   :func:`_orm.keyfunc_mapping`

"""

attribute_mapped_collection = attribute_keyed_dict
"""A synonym for :func:`_orm.attribute_keyed_dict`.

.. versionchanged:: 2.0 Renamed :data:`.attribute_mapped_collection` to
   :func:`_orm.attribute_keyed_dict`

"""

column_mapped_collection = column_keyed_dict
"""A synonym for :func:`_orm.column_keyed_dict.

.. versionchanged:: 2.0 Renamed :func:`.column_mapped_collection` to
   :func:`_orm.column_keyed_dict`

"""



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore

"""
OpenTelemetry Base Distribution (Distro)
"""

from abc import ABC, abstractmethod
from logging import getLogger

from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from opentelemetry.util._importlib_metadata import EntryPoint

_LOG = getLogger(__name__)


class BaseDistro(ABC):
    """An ABC for distro"""

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = object.__new__(cls, *args, **kwargs)

        return cls._instance

    @abstractmethod
    def _configure(self, **kwargs):
        """Configure the distribution"""

    def configure(self, **kwargs):
        """Configure the distribution"""
        self._configure(**kwargs)

    def load_instrumentor(  # pylint: disable=no-self-use
        self, entry_point: EntryPoint, **kwargs
    ):
        """Takes an instrumentation entry point and activates it by instantiating
        and calling instrument() on it.
        This is called for each opentelemetry_instrumentor entry point by auto
        instrumentation.

        Distros can override this method to customize the behavior by
        inspecting each entry point and configuring them in special ways,
        passing additional arguments, load a replacement/fork instead,
        skip loading entirely, etc.
        """
        instrumentor: BaseInstrumentor = entry_point.load()
        instrumentor().instrument(**kwargs)


class DefaultDistro(BaseDistro):
    def _configure(self, **kwargs):
        pass


__all__ = ["BaseDistro", "DefaultDistro"]



from xml.etree import cElementTree as eTree
from typing import Optional

from sysactions import format_amount


def check_change_limit_error(
    model,          # type: eTree.Element
    change,         # type: float
    change_limit,   # type: float
):
    # type: (...) ->  Optional[str]

    if change < change_limit:
        return None

    if change_limit > 0:
        change_limit_amount = format_amount(
            model=model,
            amount=change_limit,
        )
        change_amount = format_amount(
            model=model,
            amount=change,
        )
        return "$MAXIMUM_CHANGE_ALLOWED|{}|{}".format(change_limit_amount, change_amount)

    return "$AMOUNT_GREATER_THAN_DUE"



import json

from typing import Optional

from bustoken import (
    TK_ALERTS_GET_CONFIGURED_ALERTS,
    TK_ALERTS_UPDATE_ALERTS,
)
from msgbus import (
    MBException,
    TK_SYS_ACK,
    TK_SYS_NAK,
)
from sysactions import (
    action,
    close_asynch_dialog,
    get_model,
    show_any_dialog,
    show_messagebox,
    translate_message,
    StopAction,
)

from .. import mb_context
from .configured_alerts_parser import ConfiguredAlertsParser


@action
def show_configured_alerts(
    pos_id,            # type: int
    operator_id=None,  # type: Optional[int]
):
    # type: (...) -> None

    while True:
        resp = _show_configured_alerts(
            pos_id=pos_id,
            operator_id=operator_id,
        )
        if resp is None:
            raise StopAction()

        _update_alerts(
            pos_id,
        )


def _show_configured_alerts(
    pos_id,            # type: int
    operator_id=None,  # type: Optional[int]
):
    # type: (...) -> Optional[str]

    ret = None
    try:
        ret = mb_context.MB_EasySendMessage('Alerts', TK_ALERTS_GET_CONFIGURED_ALERTS)
        if ret.token == TK_SYS_NAK:
            show_messagebox(pos_id, "$GET_CONFIGURED_ALERTS_FAILED", "$ERROR")
            return
    except MBException as ex:
        if "SE_NOTFOUND" in ex.message:
            show_messagebox(
                posid=pos_id,
                message="$COMPONENT_NOT_FOUND|Alerts",
            )
            return None

    parser = ConfiguredAlertsParser(
        pos_id,
        configured_alerts=json.loads(ret.data)
    )

    table_headers = "$USER#$ALERT_TYPE#$ALERT_VALUE"
    table_body = parser.get_data_table()
    table = table_headers + "|" + "|".join(table_body)

    is_manager = not operator_id
    body = dict(isManager=is_manager, rows=table)
    options = json.dumps(body)

    last_updated = parser.get_last_updated()
    last_updated = (
        _unknow_last_updated(pos_id)
        if last_updated is None else last_updated.strftime("%x %X")
    )

    title = "$CONFIGURED_ALERTS_DIALOG_TITLE"
    message = "$CONFIGURED_ALERTS_LAST_UPDATED|{}".format(last_updated)
    buttons = "$UPDATE|$CANCEL"

    resp = show_any_dialog(
        posid=pos_id,
        dlgtype="tabularDataListDialog",
        message=message,
        title=title,
        infotext=options,
        buttons=buttons,
        icon="",
        timeout=600000,
        mask="",
        defvalue="",
        minvalue="",
        maxvalue="",
        contents=None,
        asynch=False,
    )

    return resp


def _update_alerts(
    pos_id,            # type: int
):
    # type: (...) -> None

    wait_dialog_id = show_messagebox(
        posid=pos_id,
        message="$UPDATING_CONFIGURED_ALERTS_PLEASE_WAIT",
        title="$PLEASE_WAIT",
        buttons="",
        asynch=True
    )

    message = "$CONFIGURED_ALERTS_UPDATE_FAILED"
    title = "$ERROR"

    try:
        ret = mb_context.MB_EasySendMessage('Alerts', TK_ALERTS_UPDATE_ALERTS)
        if ret.token == TK_SYS_ACK:
            message = "$CONFIGURED_ALERTS_SUCCESSFULLY_UPDATED"
            title = "$INFORMATION"
    finally:
        close_asynch_dialog(
            posid=pos_id,
            dlgid=wait_dialog_id,
        )
        show_messagebox(
            posid=pos_id,
            message=message,
            title=title,
        )


def _unknow_last_updated(
    pos_id,  # type: int
):
    # type: (...) -> str

    model = get_model(pos_id)
    return translate_message(model, "CONFIGURED_ALERTS_LAST_UPDATED_UNKNOW")



from sysactions import (
    action,
    get_posfunction,
    StopAction,
    show_messagebox,
)

from typing import (
    Any,
    Union,
)
from xml.etree import cElementTree as eTree
from mw_helper import show_message_dialog

from actions.config import (
    can_create_table_parallel_order,
)
from actions.util import check_sangria
from application.interactor import (
    CreateTableOrderInteractor,
    TableInteractor,
)
from thread_local import get_current_model


@action
def create_table_order(
    pos_id,     # type: Union[str, int]
    table_id,   # type: str
):
    # type: (...) -> Union[str, bool]

    from msgbusboundary import (
        MsgBusOrderManager,
        MsgBusTableService as tableService,
    )
    from msgbusboundary.ui import MsgBusScreenChanger

    model = get_current_model()
    check_sangria(model, pos_id)
    order_manager = MsgBusOrderManager()
    screen_changer = MsgBusScreenChanger()

    current_table_picture = eTree.XML(tableService().get_table_picture(
        pos_id=pos_id,
        table_id=table_id,
    ))
    current_pos_id = current_table_picture.get("currentPOSId", None)
    if current_pos_id is None or current_pos_id != str(pos_id):
        validate_if_can_add_order = _validate_if_can_add_order(
            pos_id=pos_id,
            table_id=table_id,
            table_picture=current_table_picture
        )
        if not validate_if_can_add_order:
            return False

    if not tableService().is_a_tab(pos_id, table_id):
        parent_table_id = table_id
        table_id = _select_table_or_related_tab_by_dialog(
            pos_id=pos_id,
            table_id=table_id,
        )

        if "TAB" in table_id and parent_table_id != table_id:
            return "0"

    create_table_order_interactor = CreateTableOrderInteractor(
        table_service=tableService(),
        screen_changer=screen_changer,
        order_manager=order_manager,
    )

    if can_create_table_parallel_order():
        tableService().store_service(
            pos_id=pos_id,
            table_id=str(table_id),
        )

    status_order = create_table_order_interactor.execute(
        pos_id=pos_id,
        table_id=table_id,
    )

    return status_order


def _select_table_or_related_tab_by_dialog(
    pos_id,     # type: Any[Union[str, int]]
    table_id,   # type: Any[Union[str, int]]
):
    # type: (...) -> Any[Union[str, int]]

    from msgbusboundary import MsgBusTableService as tableService
    from table_actions import select_table_or_tab
    from actions.table import deselect_table

    table_interactor = TableInteractor(pos_id, table_id, tableService())
    related_services, totaled_services = table_interactor.get_related_services(
        filter_totaled=True
    )

    if not related_services and totaled_services:
        show_messagebox(pos_id, "$HAS_NO_OPENED_TABS", "$INFORMATION")
        raise StopAction()

    selected_table_id = table_interactor.get_selected_related_services(related_services)
    if selected_table_id != table_id:
        deselect_table(pos_id)
        ret = select_table_or_tab(pos_id, selected_table_id)
        if ret:
            return selected_table_id
        else:
            raise StopAction()

    return selected_table_id


def _validate_if_can_add_order(
    pos_id,         # type: Union[str, int]
    table_id,       # type: str
    table_picture,  # type: eTree.Element
):
    # type: (...) -> bool

    from actions.table import validate_if_table_opened_another_pos
    from table_actions import (
        table_has_order_in_progress,
        ask_user_to_create_parallel_order,
    )

    model = get_current_model()
    pos_function = get_posfunction(model=model)

    table_is_opened_another_pos = validate_if_table_opened_another_pos(
        pos_id=pos_id,
        current_table_picture=table_picture,
    )
    if table_is_opened_another_pos:
        return False

    if table_has_order_in_progress(table_picture=table_picture):
        table_properties = table_picture.find("Properties")
        table_mode = table_properties.get("MODE") if table_properties is not None else None
        if table_mode == "tabs" or pos_function == "CS":
            message = "$TABLE_ORDER_IN_PROGRESS|{}".format(table_id)
            show_message_dialog(pos_id, "$INFORMATION", message)
            return False

        if not ask_user_to_create_parallel_order(pos_id=pos_id):
            return False

    return True



from ._AddTimeTagBoxBuilder import AddTimeTagBoxBuilder
from ._AllDoneBlinkBoxBuilder import AllDoneBlinkBoxBuilder
from ._AllItemHasAtLeastOneTagFilterBoxBuilder import AllItemHasAtLeastOneTagFilterBoxBuilder
from ._AllItemsDoneBoxBuilder import AllItemsDoneBoxBuilder
from ._AllServedItemsBoxBuilder import AllServedItemsBoxBuilder
from ._CFHHiderBoxBuilder import CFHHiderBoxBuilder
from ._CollapseSameItemsBoxBuilder import CollapseSameItemsBoxBuilder
from ._ComboDoneWhenAllDoneBoxBuilder import ComboDoneWhenAllDoneBoxBuilder
from ._ComboFilterBoxBuilder import ComboFilterBoxBuilder
from ._ComboServedWhenAllServedBoxBuilder import ComboServedWhenAllServedBoxBuilder
from ._ComboTaggedWhenAllHasTagBoxBuilder import ComboTaggedWhenAllHasTagBoxBuilder
from ._ConditionBoxBuilder import ConditionBoxBuilder
from ._CustomParamCourseBoxBuilder import CustomParamCourseBoxBuilder
from ._CustomParamPrepTimeBoxBuilder import CustomParamPrepTimeBoxBuilder
from ._CustomPropertyFilterBoxBuilder import CustomPropertyFilterBoxBuilder
from ._CustomPropertyValueFilterBoxBuilder import CustomPropertyValueFilterBoxBuilder
from ._DecreaseTimeTagBoxBuilder import DecreaseTimeTagBoxBuilder
from ._DisplayTimeFilterBoxBuilder import DisplayTimeFilterBoxBuilder
from ._FilterDefaultQuantityIngredientsBoxBuilder import FilterDefaultQuantityIngredientsBoxBuilder
from ._FilterDeletedItemsBeforeFirstStateBoxBuilder import FilterDeletedItemsBeforeFirstStateBoxBuilder
from ._FilterDeletedItemsNotSentToKitchenBoxBuilder import FilterDeletedItemsNotSentToKitchenBoxBuilder
from ._FilterOrderIfAllItemsHaveTagBoxBuilder import FilterOrderIfAllItemsHaveTagBoxBuilder
from ._FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder import \
    FilterOrderIfAtLeastOneItemHasAtLeastOneTagFilterBoxBuilder
from ._FilterOrderIfNotAllItemsHaveTagBoxBuilder import FilterOrderIfNotAllItemsHaveTagBoxBuilder
from ._FromOrderStateFilterBoxBuilder import FromOrderStateFilterBoxBuilder
from ._HideOptionBoxBuilder import HideOptionBoxBuilder
from ._ItemCustomParamsFilterBoxBuilder import ItemCustomParamsFilterBoxBuilder
from ._ItemPropertiesFilterBoxBuilder import ItemPropertiesFilterBoxBuilder
from ._ItemSequencerBoxBuilder import ItemSequencerBoxBuilder
from ._ItemSorterBoxBuilder import ItemSorterBoxBuilder
from ._JitAddTagBoxBuilder import JitAddTagBoxBuilder
from ._JitItemFilterBoxBuilder import JitItemFilterBoxBuilder
from ._JitSequencerItemFilterBoxBuilder import JitSequencerItemFilterBoxBuilder
from ._NoJitAddTagBoxBuilder import NoJitAddTagBoxBuilder
from ._NonComboFilterBoxBuilder import NonComboFilterBoxBuilder
from ._NonValidJitAddTagBoxBuilder import NonValidJitAddTagBoxBuilder
from ._NotShowOnKitchenAddTagBoxBuilder import NotShowOnKitchenAddTagBoxBuilder
from ._OrderProductCategoryFilterBoxBuilder import OrderProductCategoryFilterBoxBuilder
from ._OrderPropertiesFilterBoxBuilder import OrderPropertiesFilterBoxBuilder
from ._OrderRoundRobinBoxBuilder import OrderRoundRobinBoxBuilder
from ._OrderSequencerBoxBuilder import OrderSequencerBoxBuilder
from ._OrderSplitterBoxBuilder import OrderSplitterBoxBuilder
from ._OrderSpreaderBoxBuilder import OrderSpreaderBoxBuilder
from ._OrderStateFilterBoxBuilder import OrderStateFilterBoxBuilder
from ._OrderStateTimerBoxBuilder import OrderStateTimerBoxBuilder
from ._OrderSubTypeFilterBox import OrderSubTypeFilterBoxBuilder
from ._OrderUnsplitterBoxBuilder import OrderUnsplitterBoxBuilder
from ._OrderUnsplitterBoxBuilder import OrderUnsplitterBoxBuilder
from ._PodTypeFilterBoxBuilder import PodTypeFilterBoxBuilder
from ._OrderCostAdderBoxBuilder import OrderCostAdderBoxBuilder
from ._ProdStateFilterBoxBuilder import ProdStateFilterBoxBuilder
from ._ProdStateTimerBoxBuilder import ProdStateTimerBoxBuilder
from ._ProductCodeFilterBoxBuilder import ProductCodeFilterBoxBuilder
from ._RecalledOrderStampBoxBuilder import RecalledOrderStampBoxBuilder
from ._RemoveDeletedItemsBoxBuilder import RemoveDeletedItemsBoxBuilder
from ._RemoveDoingWhenDoneBoxBuilder import RemoveDoingWhenDoneBoxBuilder
from ._RemoveReprocessedOrdersBoxBuilder import RemoveReprocessedOrdersBoxBuilder
from ._ReprocessOrdersBoxBuilder import ReprocessOrdersBoxBuilder
from ._SaleTypeFilterBoxBuilder import SaleTypeFilterBoxBuilder
from ._SendMomentTimerBuilder import SendMomentTimerBuilder
from ._SequencerTagItemBoxBuilder import SequencerTagItemBoxBuilder
from ._ShowOnKitchenFilterBoxBuilder import ShowOnKitchenFilterBoxBuilder
from ._SkipCourseBoxBuilder import SkipCourseBoxBuilder
from ._SorterBoxBuilder import SorterBoxBuilder
from ._StampItemsViewBoxBuilder import StampItemsViewBoxBuilder
from ._StatisticBoxBuilder import StatisticBoxBuilder
from ._TagFilterBoxBuilder import TagFilterBoxBuilder
from ._TagItemsWhenProdStateBoxBuilder import TagItemsWhenProdStateBoxBuilder
from ._ViewBoxBuilder import ViewBoxBuilder
from ._AddAllItemsTagBoxBuilder import AddAllItemsTagBoxBuilder



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

GRAPHQL_DOCUMENT: Final = "graphql.document"
"""
The GraphQL document being executed.
Note: The value may be sanitized to exclude sensitive information.
"""

GRAPHQL_OPERATION_NAME: Final = "graphql.operation.name"
"""
The name of the operation being executed.
"""

GRAPHQL_OPERATION_TYPE: Final = "graphql.operation.type"
"""
The type of the operation being executed.
"""


class GraphqlOperationTypeValues(Enum):
    QUERY = "query"
    """GraphQL query."""
    MUTATION = "mutation"
    """GraphQL mutation."""
    SUBSCRIPTION = "subscription"
    """GraphQL subscription."""



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.  All rights reserved.
#
# Use of this source code is governed by a BSD-style
# license that can be found in the LICENSE file or at
# https://developers.google.com/open-source/licenses/bsd

"""DEPRECATED:  Declares the RPC service interfaces.

This module declares the abstract interfaces underlying proto2 RPC
services.  These are intended to be independent of any particular RPC
implementation, so that proto2 services can be used on top of a variety
of implementations.  Starting with version 2.3.0, RPC implementations should
not try to build on these, but should instead provide code generator plugins
which generate code specific to the particular RPC implementation.  This way
the generated code can be more appropriate for the implementation in use
and can avoid unnecessary layers of indirection.
"""

__author__ = 'petar@google.com (Petar Petrov)'

import warnings

warnings.warn(
    'google.protobuf.service module is deprecated. RPC implementations '
    'should provide code generator plugins which generate code specific to '
    'the RPC implementation. service.py will be removed in Jan 2025',
    stacklevel=2,
)

class RpcException(Exception):
  """Exception raised on failed blocking RPC method call."""
  pass


class Service(object):

  """Abstract base interface for protocol-buffer-based RPC services.

  Services themselves are abstract classes (implemented either by servers or as
  stubs), but they subclass this base interface. The methods of this
  interface can be used to call the methods of the service without knowing
  its exact type at compile time (analogous to the Message interface).
  """

  def GetDescriptor():
    """Retrieves this service's descriptor."""
    raise NotImplementedError

  def CallMethod(self, method_descriptor, rpc_controller,
                 request, done):
    """Calls a method of the service specified by method_descriptor.

    If "done" is None then the call is blocking and the response
    message will be returned directly.  Otherwise the call is asynchronous
    and "done" will later be called with the response value.

    In the blocking case, RpcException will be raised on error.

    Preconditions:

    * method_descriptor.service == GetDescriptor
    * request is of the exact same classes as returned by
      GetRequestClass(method).
    * After the call has started, the request must not be modified.
    * "rpc_controller" is of the correct type for the RPC implementation being
      used by this Service.  For stubs, the "correct type" depends on the
      RpcChannel which the stub is using.

    Postconditions:

    * "done" will be called when the method is complete.  This may be
      before CallMethod() returns or it may be at some point in the future.
    * If the RPC failed, the response value passed to "done" will be None.
      Further details about the failure can be found by querying the
      RpcController.
    """
    raise NotImplementedError

  def GetRequestClass(self, method_descriptor):
    """Returns the class of the request message for the specified method.

    CallMethod() requires that the request is of a particular subclass of
    Message. GetRequestClass() gets the default instance of this required
    type.

    Example:
      method = service.GetDescriptor().FindMethodByName("Foo")
      request = stub.GetRequestClass(method)()
      request.ParseFromString(input)
      service.CallMethod(method, request, callback)
    """
    raise NotImplementedError

  def GetResponseClass(self, method_descriptor):
    """Returns the class of the response message for the specified method.

    This method isn't really needed, as the RpcChannel's CallMethod constructs
    the response protocol message. It's provided anyway in case it is useful
    for the caller to know the response type in advance.
    """
    raise NotImplementedError


class RpcController(object):

  """An RpcController mediates a single method call.

  The primary purpose of the controller is to provide a way to manipulate
  settings specific to the RPC implementation and to find out about RPC-level
  errors. The methods provided by the RpcController interface are intended
  to be a "least common denominator" set of features which we expect all
  implementations to support.  Specific implementations may provide more
  advanced features (e.g. deadline propagation).
  """

  # Client-side methods below

  def Reset(self):
    """Resets the RpcController to its initial state.

    After the RpcController has been reset, it may be reused in
    a new call. Must not be called while an RPC is in progress.
    """
    raise NotImplementedError

  def Failed(self):
    """Returns true if the call failed.

    After a call has finished, returns true if the call failed.  The possible
    reasons for failure depend on the RPC implementation.  Failed() must not
    be called before a call has finished.  If Failed() returns true, the
    contents of the response message are undefined.
    """
    raise NotImplementedError

  def ErrorText(self):
    """If Failed is true, returns a human-readable description of the error."""
    raise NotImplementedError

  def StartCancel(self):
    """Initiate cancellation.

    Advises the RPC system that the caller desires that the RPC call be
    canceled.  The RPC system may cancel it immediately, may wait awhile and
    then cancel it, or may not even cancel the call at all.  If the call is
    canceled, the "done" callback will still be called and the RpcController
    will indicate that the call failed at that time.
    """
    raise NotImplementedError

  # Server-side methods below

  def SetFailed(self, reason):
    """Sets a failure reason.

    Causes Failed() to return true on the client side.  "reason" will be
    incorporated into the message returned by ErrorText().  If you find
    you need to return machine-readable information about failures, you
    should incorporate it into your response protocol buffer and should
    NOT call SetFailed().
    """
    raise NotImplementedError

  def IsCanceled(self):
    """Checks if the client cancelled the RPC.

    If true, indicates that the client canceled the RPC, so the server may
    as well give up on replying to it.  The server should still call the
    final "done" callback.
    """
    raise NotImplementedError

  def NotifyOnCancel(self, callback):
    """Sets a callback to invoke on cancel.

    Asks that the given callback be called when the RPC is canceled.  The
    callback will always be called exactly once.  If the RPC completes without
    being canceled, the callback will be called after completion.  If the RPC
    has already been canceled when NotifyOnCancel() is called, the callback
    will be called immediately.

    NotifyOnCancel() must be called no more than once per request.
    """
    raise NotImplementedError


class RpcChannel(object):

  """Abstract interface for an RPC channel.

  An RpcChannel represents a communication line to a service which can be used
  to call that service's methods.  The service may be running on another
  machine. Normally, you should not use an RpcChannel directly, but instead
  construct a stub {@link Service} wrapping it.  Example:

  Example:
    RpcChannel channel = rpcImpl.Channel("remotehost.example.com:1234")
    RpcController controller = rpcImpl.Controller()
    MyService service = MyService_Stub(channel)
    service.MyMethod(controller, request, callback)
  """

  def CallMethod(self, method_descriptor, rpc_controller,
                 request, response_class, done):
    """Calls the method identified by the descriptor.

    Call the given method of the remote service.  The signature of this
    procedure looks the same as Service.CallMethod(), but the requirements
    are less strict in one important way:  the request object doesn't have to
    be of any specific class as long as its descriptor is method.input_type.
    """
    raise NotImplementedError



from typing import (
    Dict,
    Any,
    List,
    Union,
)

from ._PreAccountBuilder import PreAccountBuilder
from nreports_api.shared import JSON
from table_manager_api.dto import Table
from order_api.dto import SaleLine as SaleLineApi
from nreports.services import PreAccountOrdersRetrieverService
from nreports.model import (
    GroupedSaleItems,
    PreAccountSeatTotals,
)
from .._Util import build_util_variables

LABEL_TABLE = "$TABLE"


class PreAccountBySeatsBuilder(PreAccountBuilder):

    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        pos_id = int(kwargs["posId"])
        table_id = kwargs.get("tableId")
        filter_table_seat = kwargs.get("filterTableSeat", "0")
        printed_by = kwargs.get("printBy", LABEL_TABLE)

        data_to_build = build_util_variables(
            constants=self.constants,
            data_to_build=data_to_build,
            **kwargs,
        )

        order_retriever_service = PreAccountOrdersRetrieverService(
            message_bus=self._message_bus,
            user_repository=self.user_repository,
            store_config_repository=self.store_config_repository,
        )
        table = order_retriever_service.get_table_order(
            pos_id=pos_id,
            table_id=table_id,
        )

        currency_symbol = self.i18n_repository.build_l10n_repository(pos_id=pos_id).get_currency_symbol()
        data_to_build["CURRENCY_SYMBOL"] = currency_symbol.strip()

        self._build_header(
            data_to_build=data_to_build,
            table=table,
        )

        cut_paper_by_seats = self.constants.get("PRE_ACCOUNT_CUT_PAPER_BY_SEAT", "false").lower() == "true"
        self._fill_sale_lines_by_seat(
            table=table,
            filter_table_seat=filter_table_seat,
            printed_by=printed_by,
            cut_paper_by_seats=cut_paper_by_seats,
            data_to_build=data_to_build,
        )

        data_to_build["CUT_PAPER_BEFORE_TOTALS"] = False
        if filter_table_seat == "0" and cut_paper_by_seats:
            data_to_build["CUT_PAPER_BEFORE_TOTALS"] = True

        data_to_build["SHOW_PRE_ACCOUNT_TOTALS"] = False
        if filter_table_seat == "0":
            data_to_build["SHOW_PRE_ACCOUNT_TOTALS"] = True
            self._build_totals(
                data_to_build=data_to_build,
                table=table,
            )

        self._build_footer(data_to_build=data_to_build)

        return data_to_build

    def _fill_sale_lines_by_seat(
        self,
        table: Table,
        filter_table_seat: str,
        printed_by: str,
        cut_paper_by_seats: bool,
        data_to_build: Dict[str, JSON],
    ) -> None:

        grouped_sale_lines = self._get_sale_lines_grouped_by_seats(
            table=table,
            printed_by=printed_by,
        )
        seat_quantities = self._get_seat_quantities(
            grouped_sale_lines=grouped_sale_lines,
            printed_by=printed_by,
        )
        table_sale_lines = grouped_sale_lines[LABEL_TABLE] if LABEL_TABLE in grouped_sale_lines else {}
        table_sub_total = 0.0 + self._get_sub_total_table(sale_lines=table_sale_lines)
        tip_excluded_amount = sum(table.detailed_tip_excluded.values())
        tip_percentage = self._get_percentage(
            amount=table.tip,
            sub_total_amount=(table.sub_total_amount - tip_excluded_amount),
        )
        table_tip = self._get_table_tip(
            table=table,
            table_sub_total=table_sub_total,
            tip_percentage=tip_percentage,
        )
        table_payment = table.detailed_payments[0] if 0 in table.detailed_payments.keys() else 0.0
        table_per_seat = round(table_sub_total / seat_quantities, 2)
        table_discount_per_seats = self._get_table_discount_per_seats(
            discounts_groups=table.discount_groups,
            table_sale_lines=table_sale_lines,
        )

        data_by_seats: Dict[str, JSON] = {}
        sorted_grouped_sale_lines = sorted(grouped_sale_lines, key=lambda g: (1, g) if isinstance(g, int) else (0, g))
        for index, seat in enumerate(sorted_grouped_sale_lines):
            if filter_table_seat != "0" and filter_table_seat != str(seat) and seat != "0":
                continue

            sale_lines = grouped_sale_lines[seat]
            if not sale_lines:
                continue

            is_table = seat == LABEL_TABLE
            seat_sub_total = sum([float(x.item_price) if x.item_price and x.quantity != 0 else 0.0 for x in sale_lines])

            pre_account_seat_totals = PreAccountSeatTotals(
                sale_lines=sale_lines,
                table=table,
                seat=seat,
                table_sub_total=table_sub_total,
                table_tip=table_tip,
                seat_quantities=seat_quantities,
                table_payment=table_payment,
            )
            if not is_table:
                pre_account_seat_totals.build_totals(
                    seat_sub_total=seat_sub_total,
                    seat_quantities=seat_quantities,
                    table_discount_per_seats=table_discount_per_seats,
                    table_per_seat=table_per_seat,
                    tip_percentage=tip_percentage,
                )

            data_by_seats[seat] = data_to_build["template"].copy()
            self._build_separator_seats(
                filter_table_seat=filter_table_seat == "0",
                data_by_seats=data_by_seats[seat],
                cut_paper_by_seats=cut_paper_by_seats,
            )

            data_by_seats[seat]["DIVISION_FOR_TITLE"] = self._get_division_for_title(
                seat=seat,
                is_table=is_table,
                printed_by=printed_by,
            )
            not_is_last_item = (index + 1) < len(sorted_grouped_sale_lines) and filter_table_seat == "0"
            data_by_seats[seat]["NOT_IS_LAST_ITEM"] = not_is_last_item

            self._build_conditions(
                data_by_seats=data_by_seats,
                seat=seat,
                seat_sub_total=seat_sub_total,
                table_per_seat=table_per_seat,
            )

            self._build_discount_fields(
                discount_groups=pre_account_seat_totals.discounts_groups_sale_lines,
                total_discounts=pre_account_seat_totals.total_seat_discount,
                data_to_build=data_by_seats[seat],
                list_name="LIST_DISCOUNT_SEAT",
                condition_name="HAS_DISCOUNT_LIST_SEAT"
            )

            self._build_amounts(
                data_by_seats=data_by_seats,
                seat=seat,
                is_table=is_table,
                seat_payment=pre_account_seat_totals.seat_payment,
                seat_tip=pre_account_seat_totals.seat_tip,
                seat_total=pre_account_seat_totals.seat_total,
            )

            products_lines = self._build_product_lines(sales_line_list=sale_lines)
            data_by_seats[seat]["ITEM_PRODUCT"] = products_lines

        data_to_build["saleLinesInfo"].extend(data_by_seats.values())

    @staticmethod
    def _build_separator_seats(
        filter_table_seat: bool,
        data_by_seats: Dict[str, JSON],
        cut_paper_by_seats: bool,
    ) -> None:

        data_by_seats["CUT_PAPER"] = False
        data_by_seats["SEPARATOR_DASH"] = False
        if not filter_table_seat:
            return

        if cut_paper_by_seats:
            data_by_seats["CUT_PAPER"] = True
            return

        data_by_seats["SEPARATOR_DASH"] = True

    def _get_sale_lines_grouped_by_seats(
        self,
        table: Table,
        printed_by: str,
    ) -> Dict[Union[int, str], List[SaleLineApi]]:

        grouped_sales_lines: Dict[str, List[SaleLineApi]] = {}
        grouped_sale_lines = self._get_grouped_sale_lines(table_orders=table.orders)
        fraction_line_numbers = []
        for _, sale_lines in grouped_sale_lines.items():
            fraction_line_numbers.extend(self._get_fraction_sale_line_identifiers(sales_line_list=sale_lines))

        hide_items_without_price = self.constants.get("HIDE_UNPRICED_ITEMS_FROM_PRE_ACOUNT", "true").lower() == "true"
        for seat, sale_lines in grouped_sale_lines.items():
            sale_lines_by_seat = GroupedSaleItems().grouped_sale_lines(
                sale_lines=sale_lines,
                hide_items_without_price=hide_items_without_price,
                fraction_line_numbers=fraction_line_numbers,
            )
            grouped_sales_lines.setdefault(seat, sale_lines_by_seat)

        if printed_by == "TAB":
            return {key: value for key, value in grouped_sales_lines.items() if value}

        return grouped_sales_lines

    @staticmethod
    def _get_table_tip(
        table: Table,
        table_sub_total: float,
        tip_percentage: float,
    ) -> float:

        table_tip_excluded = table.detailed_tip_excluded[0] if 0 in table.detailed_tip_excluded.keys() else 0.0
        table_tip = ((tip_percentage * (table_sub_total - table_tip_excluded)) / 100)
        return table_tip

    @staticmethod
    def _get_seat_quantities(
        grouped_sale_lines: Dict[Union[int, str], List[SaleLineApi]],
        printed_by: str,
    ) -> int:

        seat_reduction = 1 if LABEL_TABLE in grouped_sale_lines and printed_by != "TAB" else 0
        seat_quantities = len(grouped_sale_lines.keys()) - seat_reduction
        return seat_quantities

    @staticmethod
    def _build_amounts(
        data_by_seats: Dict[str, JSON],
        seat: Union[int, str],
        is_table: bool,
        seat_payment: float,
        seat_tip: float,
        seat_total: float,
    ) -> None:

        data_by_seats[seat]["HAS_PROMOTIONAL_TENDERS_SEAT"] = seat_payment > 0 and not is_table
        data_by_seats[seat]["PROMOTIONAL_TENDERS_SEAT"] = -seat_payment
        data_by_seats[seat]["HAS_TIP_AMOUNT"] = seat_tip > 0
        data_by_seats[seat]["TOTAL_WITHOUT_TIP_SEAT"] = seat_total - seat_tip
        data_by_seats[seat]["TIP_SEAT"] = seat_tip
        data_by_seats[seat]["HAS_TOTAL_SEAT"] = not is_table
        data_by_seats[seat]["TOTAL_SEAT"] = seat_total

    @staticmethod
    def _build_conditions(
        data_by_seats: Dict[str, JSON],
        seat: Union[int, str],
        seat_sub_total: float,
        table_per_seat: float,
    ) -> None:

        is_table = seat == LABEL_TABLE
        data_by_seats[seat]["IS_DIVISION_FOR_SEATS"] = table_per_seat > 0 and is_table
        data_by_seats[seat]["DIVISION_FOR_SEATS"] = table_per_seat
        data_by_seats[seat]["IS_DIVISION_FOR_TABLE"] = table_per_seat > 0 and not is_table
        data_by_seats[seat]["DIVISION_FROM_TABLE"] = table_per_seat
        data_by_seats[seat]["SUB_TOTAL_SEAT"] = seat_sub_total

    @staticmethod
    def _get_sub_total_table(
        sale_lines: List[SaleLineApi],
    ) -> float:

        return sum([float(x.item_price) if x.item_price and x.quantity != 0 else 0.0 for x in sale_lines])

    @staticmethod
    def _get_division_for_title(
        seat: str,
        is_table: bool,
        printed_by: str,
    ) -> str:

        if is_table:
            return seat

        name_by = "$SEAT_NUMBER" if printed_by == "SEATS" else "$TAB_NAME"
        return seat if is_table else f"{name_by} {seat}"

    @staticmethod
    def _get_table_discount_per_seats(
        discounts_groups: Dict[str, Dict[str, float]],
        table_sale_lines: List[SaleLineApi],
    ) -> Dict[str, float]:

        table_discount_per_seats = {}
        for table_sale_lines in table_sale_lines:
            for group in discounts_groups:
                if group not in table_discount_per_seats:
                    table_discount_per_seats[group] = 0

                line_id = "{}.{}-{}.{}".format(
                    table_sale_lines.line_number,
                    table_sale_lines.level,
                    table_sale_lines.item_id,
                    table_sale_lines.part.part_code,
                )
                table_discount_per_seats[group] = discounts_groups[group].get(line_id, 0)

        return table_discount_per_seats

    @staticmethod
    def _get_sale_lines(
        sale_lines: List[SaleLineApi],
    ) -> List[JSON]:

        item_products = []
        for sale_line in sale_lines:
            item_products.append({
                "ITEM_PRODUCT_QUANTITY": sale_line.quantity,
                "ITEM_PRODUCT_NAME": sale_line.part.product_name,
                "ITEM_PRODUCT_TOTAL_PRICE": sale_line.item_price,
            })

        return item_products

    @staticmethod
    def _get_percentage(
        amount: float,
        sub_total_amount: float,
    ) -> float:

        if amount > 0:
            return round((amount * 100) / sub_total_amount, 2)

        return 0



import re

from datetime import (
    datetime,
    timedelta,
)
from logging import Logger

from typing import (
    List,
    Any,
    Callable,
    Tuple,
    Optional,
)

from production.model import (
    Item,
    ProductionOrder, OrderChanger,
)

from ._SplitterUtil import (
    must_stay_in_main_order,
    fire_in_item,
    priority_in_item,
    hold_in_item,
    clear_fire,
    clear_priority,
    clear_hold,
)
from .._OrderChangerProductionBox import OrderChangerProductionBox
from ...manager import (
    PublishScheduler,
)


class FireHoldPriorityOrderSplitter(OrderChangerProductionBox):
    def __init__(
        self,                   # type: FireHoldPriorityOrderSplitter
        name,                   # type: str
        sons,                   # type: object
        order_changer,          # type: OrderChanger
        publish_scheduler,      # type: PublishScheduler
        logger=None,            # type: Optional[Logger]
    ):
        # type: (...) -> None

        super(FireHoldPriorityOrderSplitter, self).__init__(name, sons, logger)

        self.order_changer = order_changer
        self.publish_scheduler = publish_scheduler

    def change_order(
        self,       # type: FireHoldPriorityOrderSplitter
        order,      # type: ProductionOrder
    ):
        # type: (...) -> List[ProductionOrder]

        orders = self.split_fire_hold_priority(order)
        self._schedule_hold_orders(orders)
        self._remove_split_tags(orders)
        self._add_fire_hold_tag(orders)

        return orders

    def split_fire_hold_priority(
        self,       # type: FireHoldPriorityOrderSplitter
        order,      # type: ProductionOrder
    ):
        # type: (...) -> List[ProductionOrder]

        orders = []
        main_order_items = []
        split_index = 1
        for item in order.items:
            if not must_stay_in_main_order(item):
                main_order_items.append(item)
                continue

            hold_item, split_index = self._handle_hold_item(
                item,
                order,
                orders,
                split_index
            )

            fire_item, split_index = self._handle_fire(
                item,
                order,
                orders,
                split_index
            )

            priority_item, split_index = self._handle_priority(
                item,
                order,
                orders,
                split_index
            )

            if hold_item or fire_item or priority_item:
                property_list = [hold_in_item, fire_in_item, priority_in_item]
                clear_item = self._get_item_without_properties(
                    item=item.clone(),
                    item_has_properties=property_list,
                )
                if clear_item is not None:
                    main_order_items.append(clear_item)

        self.order_changer.change_order(order.order_id, self._mark_items_as_split)

        new_order = order
        if len(orders) > 0:
            new_order = order.clone()
            new_order.order_id = str(order.order_id) + "-0"

        new_order.items = main_order_items
        orders.insert(0, new_order)

        return orders

    def _handle_hold_item(
        self,               # type: FireHoldPriorityOrderSplitter
        item,               # type: Item
        order,              # type: ProductionOrder
        orders,             # type: List[ProductionOrder]
        split_index,        # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        stored_orders_states = [x for x in order.state_history if x.state == "STORED"]
        if len(stored_orders_states) == 0:
            return [], split_index

        hold_item_list = self._get_item_list_with_property(item.clone(), hold_in_item)
        if not hold_item_list:
            return hold_item_list, split_index

        hold_times_items = {}
        for hold_item in hold_item_list:
            property_list = [fire_in_item, priority_in_item]
            remove_tags_list = [clear_fire, clear_priority]
            hold_item = self._get_item_with_property(
                item=hold_item,
                item_has_property=hold_in_item,
                item_has_not_property=property_list,
                clear_tags=remove_tags_list,
            )
            hold_time = hold_item.properties.get("hold", 0)
            hold_time = int(hold_time)
            if hold_time not in hold_times_items:
                hold_times_items[hold_time] = []

            hold_times_items[hold_time].append(hold_item)

        for key in hold_times_items.keys():
            hold_type = "hold_" + str(key)
            for current_order in orders:
                if current_order.type == hold_type and hold_type != "hold_0":
                    new_items = order.get_cloned_items(
                        allowed_items=hold_times_items[key],
                    )
                    current_order.items.extend(new_items)
                    break
            else:
                new_hold_order = order.clone(hold_times_items[key], split_index)
                new_hold_order.type = "hold_" + str(key)
                orders.append(new_hold_order)
                split_index += 1

        return hold_item_list, split_index

    def _schedule_hold_orders(
        self,       # type: FireHoldPriorityOrderSplitter
        orders,     # type: List[ProductionOrder]
    ):
        # type: (...) -> None

        first_store_time = self._get_order_display_time(orders)
        if first_store_time is None:
            return

        datetime_now = datetime.now()
        min_hold_time = self._get_min_hold_time(
            first_store_time=first_store_time,
            orders=orders,
            datetime_now=datetime_now,
        )
        if min_hold_time is not None:
            wait_time = min_hold_time + timedelta(seconds=1) - datetime_now
            if wait_time.total_seconds() > 0:
                self.publish_scheduler.schedule_publish(orders[0].order_id, wait_time.seconds)

    def _get_min_hold_time(
        self,
        first_store_time,       # type: datetime
        orders,                 # type: List[ProductionOrder]
        datetime_now,           # type: datetime
    ):
        # type: (...) -> datetime

        min_hold_time = None
        filtered_orders = []
        for order in orders:
            hold_time = self._get_schedule_hold_time(order)
            if hold_time is None:
                filtered_orders.append(order)
                continue

            display_time = first_store_time + timedelta(seconds=hold_time)
            if datetime_now <= display_time:
                min_hold_time = display_time if min_hold_time is None or display_time < min_hold_time else min_hold_time
                continue

            if not order.display_time:
                order.display_time = datetime_now.strftime("%Y-%m-%dT%H:%M:%S")

            for item in order.items:
                self._remove_hold_from_items(item)

            filtered_orders.append(order)

        return min_hold_time

    def _remove_split_tags(
        self,       # type: FireHoldPriorityOrderSplitter
        orders,     # type: List[ProductionOrder]
    ):
        # type: (...) -> None

        for order in orders:
            if "-" not in order.order_id:
                continue

            split_id = order.order_id.split("-")[1] + "-"

            for item in order.items:
                self._remove_item_split_tags(item, split_id)

    def _add_fire_hold_tag(
        self,       # type: FireHoldPriorityOrderSplitter
        orders,     # type: List[ProductionOrder]
    ):
        # type: (...) -> None

        for order in orders:
            main_order = "-0" in order.order_id
            for item in order.items:
                fire_checker = fire_in_item if main_order else self._fire_in_item_properties
                items_with_fire = self._get_item_list_with_property(item, fire_checker)
                for fire_item in items_with_fire:
                    fire_item.add_tag("fire")

                hold_checker = hold_in_item if main_order else self._hold_in_item_properties
                items_with_hold = self._get_item_list_with_property(item, hold_checker)
                for hold_item in items_with_hold:
                    hold_item.add_tag("hold")

    def _remove_item_split_tags(
        self,           # type: FireHoldPriorityOrderSplitter
        item,           # type: Item
        split_id,       # type: str
    ):
        # type: (...) -> None

        regex_filter = "[0-9]-.*"
        new_tags = {}
        if item.tags:
            for tag in item.tags.keys():
                if not re.search(regex_filter, tag):
                    new_tags[tag] = tag
                elif split_id in tag:
                    new_tag = tag.replace(split_id, "")
                    new_tags[new_tag] = new_tag

        item.tags = new_tags

        for son in item.items:
            self._remove_item_split_tags(son, split_id)

    @staticmethod
    def _get_schedule_hold_time(
        order,      # type: ProductionOrder
    ):
        # type: (...) -> Optional[int]

        if not hasattr(order, 'type') or "hold_" not in order.type:
            return None

        hold_time = order.type.replace("hold_", "")
        if hold_time == "-1":
            a_day_seconds = 60 * 60 * 24
            return a_day_seconds

        try:
            hold_time = int(hold_time)
            return hold_time
        except ValueError:
            return None

    def _remove_hold_from_items(
        self,       # type: FireHoldPriorityOrderSplitter
        item,       # type: Item
    ):
        # type: (...) -> None

        if "hold" in item.properties:
            del item.properties["hold"]

        for son in item.items:
            self._remove_hold_from_items(son)

    @staticmethod
    def _get_order_display_time(
        orders,     # type: Optional[List[ProductionOrder]]
    ):
        # type: (...) -> Optional[datetime]

        stored_orders_states = [x for x in orders[0].state_history if x.state == "STORED"]
        if not stored_orders_states:
            return None

        first_store_state = stored_orders_states[0]
        first_store_time = first_store_state.timestamp.split(".")[0]
        first_store_time = datetime.strptime(first_store_time, "%Y-%m-%dT%H:%M:%S")

        return first_store_time

    @staticmethod
    def _get_hold_time(
        item,       # type: Item
    ):
        # type: (...) -> int

        try:
            hold_time = int(item.properties.get("hold"))
        except Exception:
            hold_time = -1

        return hold_time

    def _handle_fire(
        self,               # type: FireHoldPriorityOrderSplitter
        item,               # type: Item
        order,              # type: ProductionOrder
        orders,             # type: List[ProductionOrder]
        split_index,        # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        property_list = [hold_in_item, priority_in_item]
        remove_tags_list = [clear_hold, clear_priority]
        fire_item = self._get_item_with_property(
            item=item.clone(),
            item_has_property=fire_in_item,
            item_has_not_property=property_list,
            clear_tags=remove_tags_list,
        )
        if fire_item:
            fire_item.set_fire()
            new_fire_order = order.clone([fire_item], split_index)
            orders.append(new_fire_order)
            split_index += 1

        return fire_item, split_index

    def _handle_priority(
        self,               # type: FireHoldPriorityOrderSplitter
        item,               # type: Item
        order,              # type: ProductionOrder
        orders,             # type: List[ProductionOrder]
        split_index,        # type: int
    ):
        # type: (...) -> Tuple[Any, int]

        property_list = [fire_in_item, hold_in_item]
        remove_tags_list = [clear_hold, clear_fire]
        priority_item = self._get_item_with_property(
            item=item.clone(),
            item_has_property=priority_in_item,
            item_has_not_property=property_list,
            clear_tags=remove_tags_list,
        )
        if priority_item:
            priority_item.set_high_priority()
            new_priority_order = order.clone([priority_item], split_index)
            orders.append(new_priority_order)
            split_index += 1

        return priority_item, split_index

    def _mark_items_as_split(
        self,       # type: FireHoldPriorityOrderSplitter
        root,       # type: Any[ProductionOrder, Item]
    ):
        # type: (...) -> bool

        need_to_save_order = False
        child_need_to_save_order = False
        for item in root.items:
            need_to_save_order = self._mark_items_as_split(item)

            if "fire" in item.properties and "had_fire" not in item.tags:
                item.add_tag("had_fire")
                need_to_save_order = True

            if "hold" in item.properties and "had_hold" not in item.tags:
                item.add_tag("had_hold")
                need_to_save_order = True

            if "hold" not in item.properties and "had_hold" in item.tags and "clear_hold" not in item.tags:
                item.add_tag("clear_hold")
                need_to_save_order = True

            if "highPriority" in item.properties and "had_high_priority" not in item.tags:
                item.add_tag("had_high_priority")
                need_to_save_order = True

        return need_to_save_order or child_need_to_save_order

    def _get_item_with_property(
        self,                   # type: FireHoldPriorityOrderSplitter
        item,                   # type: Item
        item_has_property,      # type: Callable[[Item], bool]
        item_has_not_property,  # type: List[Callable]
        clear_tags,             # type: List[Callable]
    ):
        # type: (...) -> Optional[Item]

        if item_has_property(item):
            children = []
            for child in item.items or []:
                tag_child = self._get_item_without_properties(
                    item=child,
                    item_has_properties=item_has_not_property,
                )
                if tag_child:
                    children.append(tag_child)

            item.items = children
            return item

        children = []
        for child in item.items or []:
            tag_child = self._get_item_with_property(
                item=child,
                item_has_property=item_has_property,
                item_has_not_property=item_has_not_property,
                clear_tags=clear_tags,
            )
            if tag_child:
                children.append(tag_child)

        for clear_tag in clear_tags:
            clear_tag(item=item)

        item.items = children
        return item if children else None

    def _get_item_without_properties(
        self,                       # type: FireHoldPriorityOrderSplitter
        item,                       # type: Item
        item_has_properties,        # type: List[Callable[[Item], bool]]
    ):
        # type: (...) -> Optional[Item]

        if any([f(item) for f in item_has_properties]):
            return None

        if not item.items:
            return item

        children = []
        for child in item.items:
            tag_child = self._get_item_without_properties(child, item_has_properties)
            if tag_child:
                children.append(tag_child)

        item.items = children
        return item if children else None

    def _get_item_list_with_property(
        self,
        item,                               # type: Item
        item_has_property_function,         # type: Callable[[Item], Any]
        item_list_with_property=None,       # type: List
    ):
        # type: (...) -> List

        if item_list_with_property is None:
            item_list_with_property = []

        if item_has_property_function(item):
            item_list_with_property.append(item)
        else:
            for child in item.items:
                self._get_item_list_with_property(child, item_has_property_function, item_list_with_property)

        return item_list_with_property

    @staticmethod
    def _hold_in_item_properties(
        item,  # type: Item
    ):
        # type: (...) -> bool

        return "hold" in item.properties

    @staticmethod
    def _fire_in_item_properties(
        item,  # type: Item
    ):
        # type: (...) -> bool

        return "fire" in item.properties



from messagebus import (
    TokenPriority,
    TokenCreator,
)

event_manager_group = "5F"

TK_EVENT_MANAGER_PUBLISH = TokenCreator.create_token(
    priority=TokenPriority.high,
    message_group=event_manager_group,
    number="1",
)

TK_EVENT_MANAGER_GET_PENDING = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=event_manager_group,
    number="2",
)

TK_EVENT_MANAGER_GET_ABANDONED = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=event_manager_group,
    number="3",
)

TK_EVENT_MANAGER_GET_DETAILS = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=event_manager_group,
    number="4",
)

TK_EVENT_MANAGER_DEFINITION_NOT_FOUND = TokenCreator.create_token(
    priority=TokenPriority.low,
    message_group=event_manager_group,
    number="F004",
)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Final

OPENTRACING_REF_TYPE: Final = "opentracing.ref_type"
"""
Parent-child Reference type.
Note: The causal relationship between a child Span and a parent Span.
"""


class OpentracingRefTypeValues(Enum):
    CHILD_OF = "child_of"
    """The parent Span depends on the child Span in some capacity."""
    FOLLOWS_FROM = "follows_from"
    """The parent Span doesn't depend in any way on the result of the child Span."""



import json

from boss.interactor.insert_or_update_users.dto import InsertOrUpdateUsersRequest
from boss.model import User
from boss_api.shared.exception import (
    InvalidDataFormat,
    MissingMandatoryField,
    NoData,
    NoUtf8Data,
)
from ._BaseProcessor import BaseProcessor


class InsertOrUpdateUsersProcessor(BaseProcessor):

    def parse_data(
        self,
        data: bytes
    ) -> InsertOrUpdateUsersRequest:

        if not data:
            raise NoData()

        try:
            data_str = data.decode()
        except UnicodeDecodeError:
            raise NoUtf8Data()

        try:
            users_dict = json.loads(s=data_str)
            users_to_insert_or_update = []
            for user in users_dict:
                if "user_id" not in user:
                    raise MissingMandatoryField()

                users_to_insert_or_update.append(
                    User(
                        user_id=int(user["user_id"]),
                        user_name=user.get("user_name"),
                        long_name=user.get("long_name"),
                        password=user.get("password"),
                        level=int(user["level"]) if "level" in user else None,
                        status=int(user["status"]) if "status" in user else None,
                        admission_date=user.get("admission_date"),
                        termination_date=user.get("termination_date"),
                        fingerprint_minutiae=user.get("fingerprint_minutiae"),
                        nfc_uuid=user.get("nfc_uuid"),
                    )
                )
        except Exception as ex:
            exception = InvalidDataFormat()
            exception.message = str(ex)
            raise exception

        return InsertOrUpdateUsersRequest(users_to_insert_or_update=users_to_insert_or_update)



from application.domain import TableService  # noqa


class CloseTableInteractor(object):
    def __init__(self, table_service):
        # type: (TableService) -> None
        self.table_service = table_service

    def execute(
        self,           # type: CloseTableInteractor
        pos_id,         # type: str
        table_id,       # type: str
        user_id,        # type: str
    ):
        # type: (...) -> str

        return self.table_service.close_table(
            pos_id=pos_id,
            table_id=table_id,
            user_id=user_id
        )



from typing import List, Dict, Any
from pos_model import Order

from _SatXmlPartBuilder import SatXmlPartBuilder


class InfCfeBuilder(SatXmlPartBuilder):
    cnf_random = ""
    cdv = ""

    def __init__(
        self,               # type: InfCfeBuilder
        part_builders,      # type: List[SatXmlPartBuilder]
        layout_version,     # type: str
    ):
        # type: (...) -> None

        self.part_builders = part_builders
        self.layout_version = layout_version

    def build_xml(
        self,       # type: InfCfeBuilder
        order,      # type: Order
        context,    # type: Dict[unicode, Any]
    ):
        # type: (...) -> unicode

        inner_xml = ""
        for part_builder in self.part_builders:
            inner_xml += part_builder.build_xml(order, context)

        infcfe_xml = "<infCFe versaoDadosEnt='{}'>".format(self.layout_version)
        infcfe_xml += inner_xml
        infcfe_xml += "</infCFe>"

        return infcfe_xml



class CodedError(Exception):

    def __init__(
        self,       # type: CodedError
        code,       # type: int
        message,    # type: str
    ):
        # type: (...) -> None

        self.code = code
        self.message = message


class StartServiceError(object):

    POS_ID_IS_MANDATORY = CodedError(
        code=1,
        message="pos_id is mandatory",
    )


class TabServicePictureError(object):

    PARAMETER_NOT_FOUND = CodedError(
        code=-1,
        message="parameter not found",
    )

    INVALID_JSON = CodedError(
        code=-2,
        message="invalid json type for request",
    )

    REQUEST_NOT_A_DICT = CodedError(
        code=-3,
        message="input json be a dict",
    )

    EMPTY_DICT = CodedError(
        code=-4,
        message="input json dict must not be empty",
    )

    ANY_VALID_PARAM_FOUND = CodedError(
        code=-5,
        message="tab_id, tab_name or pos_id must exists",
    )

    TAB_ID_NOT_STRING = CodedError(
        code=-6,
        message="tab_id must be string",
    )

    TAB_NAME_NOT_STRING = CodedError(
        code=-7,
        message="tab_name must be string",
    )

    INVALID_TAB_NAME_FORMAT = CodedError(
        code=-8,
        message="tab_name format is invalid",
    )

    POS_ID_NOT_INT = CodedError(
        code=-9,
        message="pos_id must be int",
    )


class CreateTabFromSplitServiceError(object):

    PARSE_ERROR = CodedError(
        code=1,
        message="Error parsing SetTabNameRequest",
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opentelemetry/proto/common/v1/common.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n*opentelemetry/proto/common/v1/common.proto\x12\x1dopentelemetry.proto.common.v1"\x8c\x02\n\x08\x41nyValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x14\n\nbool_value\x18\x02 \x01(\x08H\x00\x12\x13\n\tint_value\x18\x03 \x01(\x03H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x12@\n\x0b\x61rray_value\x18\x05 \x01(\x0b\x32).opentelemetry.proto.common.v1.ArrayValueH\x00\x12\x43\n\x0ckvlist_value\x18\x06 \x01(\x0b\x32+.opentelemetry.proto.common.v1.KeyValueListH\x00\x12\x15\n\x0b\x62ytes_value\x18\x07 \x01(\x0cH\x00\x42\x07\n\x05value"E\n\nArrayValue\x12\x37\n\x06values\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.AnyValue"G\n\x0cKeyValueList\x12\x37\n\x06values\x18\x01 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue"O\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x36\n\x05value\x18\x02 \x01(\x0b\x32\'.opentelemetry.proto.common.v1.AnyValue"\x94\x01\n\x14InstrumentationScope\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12;\n\nattributes\x18\x03 \x03(\x0b\x32\'.opentelemetry.proto.common.v1.KeyValue\x12 \n\x18\x64ropped_attributes_count\x18\x04 \x01(\rB{\n io.opentelemetry.proto.common.v1B\x0b\x43ommonProtoP\x01Z(go.opentelemetry.io/proto/otlp/common/v1\xaa\x02\x1dOpenTelemetry.Proto.Common.V1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "opentelemetry.proto.common.v1.common_pb2", _globals
)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals["DESCRIPTOR"]._loaded_options = None
    _globals["DESCRIPTOR"]._serialized_options = (
        b"\n io.opentelemetry.proto.common.v1B\013CommonProtoP\001Z(go.opentelemetry.io/proto/otlp/common/v1\252\002\035OpenTelemetry.Proto.Common.V1"
    )
    _globals["_ANYVALUE"]._serialized_start = 78
    _globals["_ANYVALUE"]._serialized_end = 346
    _globals["_ARRAYVALUE"]._serialized_start = 348
    _globals["_ARRAYVALUE"]._serialized_end = 417
    _globals["_KEYVALUELIST"]._serialized_start = 419
    _globals["_KEYVALUELIST"]._serialized_end = 490
    _globals["_KEYVALUE"]._serialized_start = 492
    _globals["_KEYVALUE"]._serialized_end = 571
    _globals["_INSTRUMENTATIONSCOPE"]._serialized_start = 574
    _globals["_INSTRUMENTATIONSCOPE"]._serialized_end = 722
# @@protoc_insertion_point(module_scope)
